/*====================================+=====================================+
! File CDlgInstall.h                  ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#pragma once
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
class CInstallEvent : public wxCommandEvent
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CInstallEvent( wxEventType commandType = wxEVT_NULL, int winid = 0 )
                   : wxCommandEvent( commandType, winid )
      {  ; }

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      int m_i_nb_file = 0 ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      virtual wxEvent *Clone() const override
      {  return new CInstallEvent( *this ) ; }
      /*-------------------------------------------------------------------*/
      int get_id() const { return( GetId() ) ; }
      void set_id( int i_id ) { SetId( i_id ) ; }
      int get_err_code() const { return( GetInt() ) ; }
      void set_err_code( int i_err_code ) { SetInt( i_err_code ) ; }
      /*-------------------------------------------------------------------*/
      int get_num_file() const { return( GetExtraLong() ) ; }
      void set_num_file( int i_num_file ) { SetExtraLong( i_num_file ) ; }
      int get_nb_file() const { return( m_i_nb_file ) ; }
      void set_nb_file( int i_nb_file ) { m_i_nb_file = i_nb_file ; }

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
wxDECLARE_EVENT( EVT_INSTALL, CInstallEvent ) ;

/*-------------------------------------------------------------------------*/
class CDlgInstall ;

/*-------------------------------------------------------------------------*/
class CInstallThread : public wxThread
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      explicit CInstallThread( CDlgInstall *p_inst ) ;
      ~CInstallThread() ;

      /*-------------------------------------------------------------------*/
      enum e_err_code
      {  IE_NO_ERROR_DONE, IE_NO_ERROR_INFO, IE_FILE_NOT_FOUND,
         IE_INVALID_ZIP, IE_UNEXPECTED_CONTENT, IE_READ_ERROR,
         IE_STOP_REQUEST, IE_WRITE_ERROR, IE_NB
      } ;

      /*-------------------------------------------------------------------*/
      static constexpr const char *m_st_tb_err_msg[ IE_NB ]
      = { wxTRANSLATE( "Installation done" ),          // IE_NO_ERROR_DONE
          wxTRANSLATE( "Information" ),                // IE_NO_ERROR_INFO
          wxTRANSLATE( "File not found" ),             // IE_FILE_NOT_FOUND
          wxTRANSLATE( "Invalid zip file" ),           // IE_INVALID_ZIP
          wxTRANSLATE( "Unexpected zip content" ),   // IE_UNEXPECTED_CONTENT
          wxTRANSLATE( "Read error" ),                 // IE_READ_ERROR
          wxTRANSLATE( "Installation cancelled" ),     // IE_STOP_REQUEST
          wxTRANSLATE( "Error writing to local file" ) // IE_WRITE_ERROR
        } ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      CDlgInstall *m_p_inst ; // Parent object
      wxFileName  m_fn_source ;
      /*-------------------------------------------------------------------*/
      e_err_code m_err_code = IE_NO_ERROR_DONE ;
      int        m_i_nb_file = 0 ;
      int        m_i_num_file = 0 ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      e_err_code get_err_code() const { return( m_err_code ) ; }
      static wxString st_get_err_msg( int i_err_code )
      {  return( i_err_code < 0 || i_err_code >= IE_NB
                 ? "invalid code"
                 : wxGetTranslation( m_st_tb_err_msg[ i_err_code ] )
               ) ;
      }
      /*-------------------------------------------------------------------*/
      void reset_p_inst() { m_p_inst = nullptr ; }
      e_err_code install() ;
      /*--( Will be executed by "Run" )------------------------------------*/
      virtual wxThread::ExitCode Entry() override ;
      void send_event() ;

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
class CDlgInstall : public CDlgProgress
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CDlgInstall( wxWindow *parent, const wxFileName &fn_source ) ;
      ~CDlgInstall() ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      wxFileName m_fn_source ;
      wxString   m_s_info_prev ;
      /*-------------------------------------------------------------------*/
      CInstallThread    *m_p_thread = nullptr ;
      wxCriticalSection m_csec_p_thread ;
      bool              m_boo_thread_stopping = false ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      void OnBtCancel( wxCommandEvent &event ) override ;
      void OnEvent( CInstallEvent &event ) ;
      void msg_info( const wxString &s ) ;
      void msg_error( const wxString &s ) ;
      CInstallThread::e_err_code non_threaded_exec() ;
      void thread_start() ;
      void thread_stop() ;
      bool thread_running() ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      int go() ;

   /*----------------------------------------------------------------------*/
   friend class CInstallThread ;
   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                       End of file CDlgInstall.h                           !
+==========================================================================*/
