/*====================================+=====================================+
! File CVlcMedia.h                    ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#pragma once
/*-------------------------------------------------------------------------*/
#include "CVlcInterface.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
wxDECLARE_EVENT( EVT_SET_MESSAGE, wxCommandEvent ) ;
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
class CVlcMedia : public wxWindow
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CVlcMedia( wxWindow *parent, const wxString &s_channel ) ;
      ~CVlcMedia() ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      wxPanel        *m_panel_box      = nullptr ;
      wxBoxSizer     *m_bsz_box        = nullptr ;
      wxBoxSizer     *m_bsz_gen        = nullptr ;
      wxBoxSizer     *m_bsz_info       = nullptr ;
      wxStaticBitmap *m_stb_logo       = nullptr ;
      wxStaticText   *m_stx_msg        = nullptr ;
      wxGauge        *m_ga_retry_timer = nullptr ;

      /*--------------------------------------------------------------------+
      ! One VLC instance is created by player :                             !
      ! - The result seems more "smooth" and "stable" rather only one vlc   !
      !   instance for all players                                          !
      ! - It enables to have different vlc "init strings"                   !
      +--------------------------------------------------------------------*/
      CVlcInterface *m_vlc = nullptr ;
      libvlc_media_player_t *m_mp = nullptr ;
      /*--------------------------------------------------------------------+
      ! Channel currently played : The pointer to it is not stored because  !
      ! the map it is associated to might be reallocated. For example if    !
      ! new channels are added by the user.                                 !
      +--------------------------------------------------------------------*/
      wxString m_s_channel ;
      /*--( Playlist from which the channel has been started )-------------*/
      wxString m_s_pl_name ;
      /*--( Popup menu displayed ? to avoid removing mouse pointer )-------*/
      bool m_boo_popup_menu_on = false ;
      /*--( Check / Force sound, video etc. )------------------------------*/
      wxTimer m_timer_init_audio     ;
      int     m_i_num_init_audio = 0 ;
      wxTimer m_timer_init_video     ;
      int     m_i_num_init_video = 0 ;
      /*--( Waiting for the channel to become on line )--------------------*/
      bool m_boo_play_retry = false ; // In "retry" mode ?
      int  m_i_wait_sec_retry ; // Number of seconds since last retry
      wxTimer m_timer_retry_play ;
      /*-------------------------------------------------------------------*/
      bool     m_boo_rendering_started = false ;
      /*--( Wait a bit before displaying the message )---------------------*/
      wxTimer  m_timer_message ;
      wxString m_s_message ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      void OnKeyDown( wxKeyEvent &event ) ;
      void OnMouseLeftDown( wxMouseEvent &event ) ;
      void OnMouseLeftDClick( wxMouseEvent &event ) ;
      void OnMouseMiddleDown( wxMouseEvent &event ) ;
      void OnMouseRightDown( wxMouseEvent &event ) ;
      void OnPlayRetry( wxTimerEvent &event ) ;
      /*--( Used to bind the mouse events )--------------------------------*/
      using t_mouse_event_type = const wxEventTypeTag< wxMouseEvent > & ;
      using t_mouse_event_func = void ( CVlcMedia::* )( wxMouseEvent &event);
      void bind_mouse_event( t_mouse_event_type eventType,
                             t_mouse_event_func functor
                           ) ;
      /*-------------------------------------------------------------------*/
      bool is_ok() const
      {  return( m_vlc != nullptr && m_vlc->is_ok() ) ; }
      wxWindow *get_window()
      {  return( m_panel_box ) ; }
      CVlcInterface *get_vlc_int() const
      {  return( m_vlc ) ; }
      wxString get_channel_name() const
      {  return( m_s_channel ) ; }
      CChannel *get_channel_played() const
      {  return( wxGetApp().get_channel( m_s_channel ) ) ; }
      libvlc_event_manager_t *get_media_player_evt_man()
      {  return( m_vlc == nullptr
                 ? nullptr
                 : m_vlc->m_libvlc_media_player_event_manager( m_mp )
               ) ;
      }
      libvlc_event_manager_t *get_media_evt_man( libvlc_media_t *p_md )
      {  return( m_vlc == nullptr
                 ? nullptr
                 : m_vlc->m_libvlc_media_event_manager( p_md )
               ) ;
      }
      void add_vlc_option( const wxString &s_option )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_media_add_option( get_media(), s_option ) ; }
      }
      void create_vlc_interface() ;
      void delete_vlc_interface() ;
      bool all_evt_mp_attach() ;
      void init_player_window() ;
      void reset_player_window() ;
      void init_channel( const wxString &s_channel ) ;
      void refresh_pl_channel() ;
      void manage_event( wxCommandEvent &event ) ;
      void play()
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_media_player_play( m_mp ) ; }
      }
      void toggle_pause()
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_media_player_pause( m_mp ) ; }
      }
      void stop()
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_media_player_stop( m_mp ) ; }
      }
      int get_all_track_count()
      {  return( m_vlc == nullptr
                 ? -1
                 :   m_vlc->m_libvlc_video_get_track_count( m_mp )
                   + m_vlc->m_libvlc_audio_get_track_count( m_mp )
                   + m_vlc->m_libvlc_video_get_spu_count( m_mp )
               ) ;
      }
      int get_audio_track_count()
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_audio_get_track_count( m_mp )
               ) ;
      }
      int get_video_track_count()
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_video_get_track_count( m_mp )
               ) ;
      }
      void refresh_streams( CChannel *p_channel ) ;
      void start_channel() ;
      void init_retry_channel() ;
      void stop_channel() ;
      void show_played_channel_item() ;
      libvlc_media_t *get_media() const
      {  return( m_vlc == nullptr
                 ? nullptr : m_vlc->m_libvlc_media_player_get_media( m_mp )
               ) ;
      }
      void set_mute( bool boo_mute )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_audio_set_mute( m_mp, boo_mute ) ; }
      }
      bool is_media_radio() ;
      void init_audio() ;
      void init_video() ;
      void set_media_by_location( const wxString &s_url ) ;
      libvlc_state_t get_state() const
      {  return( m_vlc == nullptr
                 ? libvlc_Error
                 : m_vlc->m_libvlc_media_player_get_state( m_mp )
               ) ;
      }
      bool is_running() const
      {  libvlc_state_t state = get_state() ;
         return(    state == libvlc_Playing
                 || state == libvlc_Paused
                 || get_play_retry()
               ) ;
      }
      bool is_playing() const
      {  return( get_state() == libvlc_Playing ) ; }
      void set_volume( int i_val )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_audio_set_volume( m_mp, i_val ) ; }
      }
      bool get_mute() const
      {  return(    m_vlc != nullptr
                 && m_vlc->m_libvlc_audio_get_mute( m_mp ) == true
               ) ;
      }
      void take_snapshot( const char *co_p_c_path )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_video_take_snapshot( m_mp,0, co_p_c_path,0,0 ) ;}
      }
      void set_video_mouse_input( bool boo_on )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_video_set_mouse_input( m_mp, boo_on ) ; }
      }
      void set_deinterlace( const char *p_c_mode )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_video_set_deinterlace( m_mp, p_c_mode ) ; }
      }
      void set_deinterlace()
      {  if( m_vlc != nullptr )
         {  const char *p_c_mode = wxGetApp().M_s_deinterlace.get() ;
            set_deinterlace( *p_c_mode == '\0' ? nullptr : p_c_mode ) ;
         }
      }
      void set_video_key_input( bool boo_on )
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_video_set_key_input( m_mp, boo_on ) ; }
      }
      /*-------------------------------------------------------------------*/
      libvlc_track_description_t *get_video_track_description() const
      {  return( m_vlc == nullptr
                 ? nullptr
                 : m_vlc->m_libvlc_video_get_track_description( m_mp )
               ) ;
      }
      int get_video_track() const
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_video_get_track( m_mp )
               ) ;
      }
      int set_video_track( int i_track )
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_video_set_track( m_mp, i_track )
               ) ;
      }
      /*-------------------------------------------------------------------*/
      libvlc_track_description_t *get_audio_track_description() const
      {  return( m_vlc == nullptr
                 ? nullptr
                 : m_vlc->m_libvlc_audio_get_track_description( m_mp )
               ) ;
      }
      int get_audio_track() const
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_audio_get_track( m_mp )
               ) ;
      }
      int set_audio_track( int i_track )
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_audio_set_track( m_mp, i_track )
               ) ;
      }
      /*-------------------------------------------------------------------*/
      libvlc_track_description_t *get_video_spu_description() const
      {  return( m_vlc == nullptr
                 ? nullptr
                 : m_vlc->m_libvlc_video_get_spu_description( m_mp )
               ) ;
      }
      int get_video_spu() const
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_video_get_spu( m_mp )
               ) ;
      }
      int set_video_spu( int i_spu )
      {  return( m_vlc == nullptr
                 ? -1 : m_vlc->m_libvlc_video_set_spu( m_mp, i_spu )
               ) ;
      }
      /*-------------------------------------------------------------------*/
      void release_track_description_list( libvlc_track_description_t *p_li
                                         ) const
      {  if( m_vlc != nullptr )
         {  m_vlc->m_libvlc_track_description_list_release( p_li ) ; }
      }
      /*-------------------------------------------------------------------*/
      bool get_rendering_started() const
      {  return( m_boo_rendering_started ) ; }
      void set_rendering_started( bool boo_val )
      {  m_boo_rendering_started = boo_val ; }
      void clear_message() ;
      void set_message_nowait( const wxString &s_msg ) ;
      void set_message( const wxString &s_msg ) ;
      void show_message() ;
      /*-------------------------------------------------------------------*/
      bool get_popup_menu_on() const { return( m_boo_popup_menu_on ) ; }
      bool get_play_retry() const { return( m_boo_play_retry ) ; }
      /*-------------------------------------------------------------------*/
      void show_focus() ;
      void change_media_active( int i_key_code ) ;

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
using t_vec_vlc_media     = std::vector< CVlcMedia * >      ;
using t_vec_vlc_media_it  = t_vec_vlc_media::iterator       ;
using t_vec_vlc_media_cit = t_vec_vlc_media::const_iterator ;
/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                         End of file CVlcMedia.h                           !
+==========================================================================*/
