/*====================================+=====================================+
! File CBookPlaylist.cpp              ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/tokenzr.h>
#include <wx/imaglist.h>
#include "CFrame.h"
#include "CDndChannel.h"
#include "CDlgPlName.h"
#include "tango/accessories_text_editor_16x16.xpm"
#include "tango/preferences_desktop_font_16x16.xpm"
#include "tango/go_up_16x16.xpm"
#include "tango/go_down_16x16.xpm"
#include "tango/media_playback_start_16x16.xpm"
#include "tango/list_add_16x16.xpm"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
CBookPlaylist::CBookPlaylist( wxWindow *parent, const wxString &s_name )
              : wxTreeCtrl( parent, wxID_ANY,
                            wxDefaultPosition, wxDefaultSize,
                            wxTR_DEFAULT_STYLE | wxTR_SINGLE | wxTR_HIDE_ROOT
                          ),
                m_panel_pl( new wxPanel( parent, wxID_ANY ) ),
                m_sizer_pl( new wxBoxSizer( wxVERTICAL ) )
{
   /*--( Associated window where the playlist is displayed )---------------*/
   m_panel_pl->SetSizer( m_sizer_pl ) ;
   m_panel_pl->SetName( "panel_pl" ) ;
   /*----------------------------------------------------------------------*/
   SetName( s_name ) ;
   AddRoot( "Root" ) ;
   SetDropTarget( new CChannelDropTargetOnTree( this ) ) ;
   SetFont( GetFont().Larger() ) ;
   /*----------------------------------------------------------------------*/
   m_il_pl = new wxImageList( 24, 24 ) ;
   AssignImageList( m_il_pl ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_TREE_ITEM_RIGHT_CLICK,
         [ this ]( wxTreeEvent &event ) { context_menu( event ) ; }
       ) ;
   Bind( wxEVT_TREE_SEL_CHANGED, &CBookPlaylist::OnTreeSelChanged, this ) ;
   Bind( wxEVT_TREE_BEGIN_DRAG, &CBookPlaylist::OnTreeBeginDrag, this ) ;
   Bind( wxEVT_TREE_END_DRAG, &CBookPlaylist::OnTreeEndDrag, this ) ;
   Bind( wxEVT_TREE_KEY_DOWN, &CBookPlaylist::OnTreeKeyDown, this ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CBookPlaylist::~CBookPlaylist()
{
   /*--( On MSW some events seem to be generated after object deletion (?))*/
   Unbind( wxEVT_TREE_SEL_CHANGED, &CBookPlaylist::OnTreeSelChanged, this ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::OnTreeSelChanged( wxTreeEvent &event )
{
   /*----------------------------------------------------------------------*/
   CPlaylist *pl_new = get_playlist( event.GetItem() ) ;
   /*----------------------------------------------------------------------*/
   if( event.GetOldItem() == event.GetItem() || pl_new == nullptr )
   {  return ; }
   /*----------------------------------------------------------------------*/
   change_window_pl( pl_new->get_main_window() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::OnTreeBeginDrag( wxTreeEvent &event )
{
   /*--( Only User playlist can be dragged )-------------------------------*/
   if( !is_parent_user_node( event.GetItem() ) ) { return ; }
   /*--( By default it is vetoed )-----------------------------------------*/
   event.Allow() ;
   m_ti_drag = event.GetItem() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::OnTreeEndDrag( wxTreeEvent &event )
{
   /*----------------------------------------------------------------------*/
   if( !m_ti_drag.IsOk() ) { return ; }
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti_dst = event.GetItem() ;
   /*----------------------------------------------------------------------*/
   if( ti_dst == m_ti_drag ) { return ; }
   /*--( User playlist can only be dragged into User list )----------------*/
   if( !is_parent_user_node( ti_dst ) ) { return ; }
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti_first = get_first_ti() ;
   /*--( Find out if the drop will be done above or below )----------------*/
   wxRect rect_dst ;
   if( !GetBoundingRect( ti_dst, rect_dst ) ) { return ; }
   /*--( Move before or after this item )----------------------------------*/
   bool boo_move_above
      = ( event.GetPoint().y < rect_dst.y + rect_dst.GetHeight() / 2 ) ;
   /*--( The "Delete" deletes the item data )------------------------------*/
   int i_il_image = GetItemImage( m_ti_drag ) ;
   wxString s_text = GetItemText( m_ti_drag ) ;
   wxTreeItemData *p_tid = new CTreeItemDataPl( get_playlist( m_ti_drag ) ) ;
   /*--( The source can now be deleted )-----------------------------------*/
   Delete( m_ti_drag ) ;
   m_ti_drag.Unset() ;
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti_new ;
   wxTreeItemId ti_parent = GetItemParent( ti_dst ) ;
   if( boo_move_above && ti_dst == ti_first )
   {  ti_new = PrependItem( ti_parent, s_text, i_il_image, -1, p_tid ) ; }
   else
   {  /*-------------------------------------------------------------------*/
      wxTreeItemId ti_prev
                 = boo_move_above ? GetPrevSibling( ti_dst ) : ti_dst ;
      /*-------------------------------------------------------------------*/
        ti_new
      = InsertItem( ti_parent, ti_prev, s_text, i_il_image, -1, p_tid ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   set_selection( ti_new ) ;
   init_list_parents( ti_new ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::OnTreeKeyDown( wxTreeEvent &event )
{
   /*----------------------------------------------------------------------*/
   const wxTreeItemId &ti = GetSelection() ;
   bool boo_cmd_key
     = ( ( event.GetKeyEvent().GetModifiers() & ~wxMOD_ALT ) == wxMOD_CMD ) ;
   /*----------------------------------------------------------------------*/
   if( event.GetKeyCode() == WXK_F2 )
   {  rename_pl( ti ) ; }
   else
   if( event.GetKeyCode() == WXK_DELETE )
   {  delete_pl( ti ) ; }
   else
   if( boo_cmd_key && event.GetKeyCode() == 'V' )
   {  get_playlist( ti )->paste_from_clipboard() ; }
   else
   {  event.Skip() ; }
   /*----------------------------------------------------------------------*/
}

/*--( Recursive function )-------------------------------------------------*/
wxTreeItemId CBookPlaylist::find_ti_name( const wxString &s_pl_name,
                                          wxTreeItemId ti_start
                                        )
{
   /*--( Default to root )-------------------------------------------------*/
   if( !ti_start.IsOk() ) { ti_start = GetRootItem() ; }
   /*--( The right element ? )---------------------------------------------*/
   CPlaylist *pl_start = get_playlist( ti_start ) ;
   if( pl_start != nullptr && pl_start->get_name() == s_pl_name )
   {  return( ti_start ) ; }
   /*--( No, scan the children )-------------------------------------------*/
   if( ItemHasChildren( ti_start ) )
   {
      /*-------------------------------------------------------------------*/
      wxTreeItemIdValue tiv_cookie ;
      wxTreeItemId ti_sub ;
      /*-------------------------------------------------------------------*/
      for( auto ti = GetFirstChild( ti_start, tiv_cookie ) ;
           ti.IsOk() ;
           ti = GetNextChild( ti_start, tiv_cookie )
         )
      {  /*--( Searching sub-tree )----------------------------------------*/
         if( ( ti_sub = find_ti_name( s_pl_name, ti ) ).IsOk() )
         {  return( ti_sub ) ; }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*--( Not found )-------------------------------------------------------*/
   return( wxTreeItemId() ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Returns a vector with the "children" playlists )---------------------*/
int CBookPlaylist::get_pl_children( e_pl_known pl_known, t_vec_pl &vec_pl )
{
   /*----------------------------------------------------------------------*/
   vec_pl.clear() ;
   /*----------------------------------------------------------------------*/
   CPlaylist *pl = wxGetApp().get_pk_pl( pl_known ) ;
   if( pl == nullptr ) { return( -1 ) ; }
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti_main = pl->get_ti() ;
   if( !ti_main.IsOk() ) { return( -2 ) ; }
   /*----------------------------------------------------------------------*/
   wxTreeItemIdValue tiv_cookie ;
   /*----------------------------------------------------------------------*/
   for( auto ti = GetFirstChild( ti_main, tiv_cookie ) ;
        ti.IsOk() ;
        ti = GetNextChild( ti_main, tiv_cookie )
      )
   {  vec_pl.emplace_back( get_playlist( ti ) ) ; }
   /*----------------------------------------------------------------------*/
   return( vec_pl.size() ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Returns a string containing the user pl names in the right order )---*/
wxString CBookPlaylist::get_pl_user_list()
{
   /*----------------------------------------------------------------------*/
   t_vec_pl vec_pl_user ;
   wxString s_pl_user ;
   /*----------------------------------------------------------------------*/
   get_pl_children( PK_USER_NODE, vec_pl_user ) ;
   /*----------------------------------------------------------------------*/
   for( const auto pl : vec_pl_user )
   {  /*--------------------------------------------------------------------+
      ! playlist names are separated with LF because user playlist names    !
      ! can contain spaces                                                  !
      +--------------------------------------------------------------------*/
      if( !s_pl_user.empty() ) { s_pl_user += '\n' ; }
      /*--( To have "real" names in the config file (cf search) )----------*/
      s_pl_user += pl->get_name() ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( s_pl_user ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Delete channels from all playlists that can have them )--------------*/
int CBookPlaylist::del_channel_pl( const mtf::t_vec_string &vs_channel )
{
   /*----------------------------------------------------------------------*/
   t_vec_pl vec_pl ;
   /*----------------------------------------------------------------------*/
   if( get_pl_children( PK_USER_NODE, vec_pl ) > 0 )
   {
      /*--( Delete channels )----------------------------------------------*/
      for( auto pl : vec_pl ) { pl->del_channel( vs_channel ) ; }
      /*--( Reinit User node and parents )---------------------------------*/
      wxGetApp().get_pk_pl( PK_USER_NODE )->init_list() ;
      /*-------------------------------------------------------------------*/
   }
   /*--( Search may list the some deleted channels )-----------------------*/
   wxGetApp().get_pk_pl( PK_SEARCH )->del_channel( vs_channel ) ;
   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Returns a string containing the expanded nodes                            !
! NOTE : - node names are unique to the entire tree !!!                     !
!        - all expanded nodes are saved even if their parents are not       !
!          this enables to keep expanded the sub nodes                      !
!        - recursive function                                               !
+--------------------------------------------------------------------------*/
wxString CBookPlaylist::get_expanded( wxTreeItemId ti_start )
{
   /*--( Default to root )-------------------------------------------------*/
   if( !ti_start.IsOk() ) { ti_start = GetRootItem() ; }
   /*----------------------------------------------------------------------*/
   wxString s_expanded ;
   wxString s_expanded_sub ;
   wxTreeItemIdValue tiv_cookie ;
   /*--( If here then only this node is expanded )-------------------------*/
   CPlaylist *pl = get_playlist( ti_start ) ;
   if( pl != nullptr && IsExpanded( ti_start ) )
   {  s_expanded = get_playlist( ti_start )->get_name() ; }
   /*----------------------------------------------------------------------*/
   for( auto ti = GetFirstChild( ti_start, tiv_cookie ) ;
        ti.IsOk() ;
        ti = GetNextChild( ti_start, tiv_cookie )
      )
   {  /*--------------------------------------------------------------------+
      ! playlist names are separated with LF because user playlist names    !
      ! can contain spaces                                                  !
      +--------------------------------------------------------------------*/
      s_expanded_sub = get_expanded( ti ) ;
      if( !s_expanded_sub.empty() )
      {  /*----------------------------------------------------------------*/
         if( !s_expanded.empty() ) { s_expanded += '\n' ; }
         /*--( One more sub node is expanded )-----------------------------*/
         s_expanded += s_expanded_sub ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( s_expanded ) ;
   /*----------------------------------------------------------------------*/
}

/*--( NOTE : node names are unique to the entire tree )--------------------*/
void CBookPlaylist::expand_list( const wxString &s_names )
{
   /*----------------------------------------------------------------------*/
   if( s_names.empty() ) { return ; }
   /*----------------------------------------------------------------------*/
   wxStringTokenizer tkz( s_names, '\n' ) ;
   /*----------------------------------------------------------------------*/
   while( tkz.HasMoreTokens() )
   {  /*-------------------------------------------------------------------*/
      wxTreeItemId ti = find_ti_name( tkz.GetNextToken() ) ;
      if( ti.IsOk() ) { Expand( ti ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CBookPlaylist::swap_ti_values( const wxTreeItemId &ti1,
                                    const wxTreeItemId &ti2
                                  )
{
   /*----------------------------------------------------------------------*/
   if( !ti1.IsOk() || !ti2.IsOk() || ti1 == ti2 ) { return( false ) ; }
   /*----------------------------------------------------------------------*/
   wxString s_text = GetItemText( ti1 ) ;
   CTreeItemDataPl *p_tidpl = get_tidpl( ti1 ) ;
   /*----------------------------------------------------------------------*/
   SetItemText( ti1, GetItemText( ti2 ) ) ;
   SetItemData( ti1, GetItemData( ti2 ) ) ;
   SetItemText( ti2, s_text ) ;
   SetItemData( ti2, p_tidpl ) ;
   /*--( ti1 anf ti2 should have the same parent )-------------------------*/
   init_list_parents( ti1 ) ;
   /*----------------------------------------------------------------------*/
   return( true ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::move_up( const wxTreeItemId &ti )
{
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti_up = GetPrevSibling( ti ) ;
   if( ti_up.IsOk() && swap_ti_values( ti, ti_up ) && IsSelected( ti ) )
   {  SelectItem( ti_up ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::move_down( const wxTreeItemId &ti )
{
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti_down = GetNextSibling( ti ) ;
   if( ti_down.IsOk() && swap_ti_values( ti, ti_down ) && IsSelected( ti ) )
   {  SelectItem( ti_down ) ; }
   /*----------------------------------------------------------------------*/
}

/*--( sz_pos is the insertion point (-1 is supposed to mean "append") )----*/
wxTreeItemId CBookPlaylist::add_pl( const wxTreeItemId &ti_parent,
                                    size_t sz_pos, CPlaylist *pl
                                  )
{
   /*----------------------------------------------------------------------*/
   if( pl != nullptr && pl->get_main_window() != nullptr )
   {  pl->get_main_window()->Show( false ) ; }
   /*----------------------------------------------------------------------*/
   pl->set_ti( InsertItem( ti_parent, sz_pos,
                           pl->get_label(), pl->get_image_il_pos(), -1,
                           new CTreeItemDataPl( pl )
                         )
             ) ;
   /*----------------------------------------------------------------------*/
   return( pl->get_ti() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxTreeItemId CBookPlaylist::add_pl( CPlaylist *pl, size_t sz_pos )
{
   /*----------------------------------------------------------------------*/
   return( add_pl( make_path( pl->get_tree_path() ), sz_pos, pl ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::rename_pl( const wxTreeItemId &ti )
{
   /*----------------------------------------------------------------------*/
   CPlaylist *pl = get_playlist( ti ) ;
   if( pl == nullptr || !pl->get_renamable() ) { return ; }
   /*----------------------------------------------------------------------*/
   wxString s_name ;
   CDlgPlName( _( "Rename Playlist" ), pl ).get_modal( s_name ) ;
   /*--( If the returned name is empty, this means an error occurred )-----*/
   if( !s_name.empty() )
   {  /*-------------------------------------------------------------------*/
      pl->rename( s_name ) ;
      SetItemText( ti, s_name ) ;
      SetItemImage( ti, pl->get_image_il_pos() ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Change the window supposed to be the information associated to the        !
! current tree item                                                         !
+--------------------------------------------------------------------------*/
void CBookPlaylist::change_window_pl( wxWindow *win_new )
{
   /*----------------------------------------------------------------------*/
   m_panel_pl->Freeze() ;
   /*----------------------------------------------------------------------*/
   if( m_sizer_pl->GetItemCount() > 0 )
   {  /*-------------------------------------------------------------------*/
      wxWindow *win_old = m_sizer_pl->GetItem( static_cast< size_t >( 0 )
                                             )->GetWindow() ;
      win_old->Show( false ) ;
      m_sizer_pl->Detach( 0 ) ;
      win_old->Reparent( GetParent() ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( win_new != nullptr )
   {  /*-------------------------------------------------------------------*/
      win_new->Reparent( m_panel_pl ) ;
      m_sizer_pl->Add( win_new, 1, wxEXPAND ) ;
      win_new->Show() ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   m_panel_pl->Layout() ;
   m_panel_pl->Thaw() ;
   /*----------------------------------------------------------------------*/
}

/*--( Just remove the playlist from the book : it is not deleted )---------*/
void CBookPlaylist::remove_pl( const wxTreeItemId &ti )
{
   /*----------------------------------------------------------------------*/
   if( !ti.IsOk() ) { return ; }
   /*--( The delete won't set a new selection automatically )--------------*/
   wxTreeItemId ti_sel = GetNextSibling( ti ) ;
   if( !ti_sel.IsOk() )
   {  ti_sel = GetPrevSibling( ti ) ;
      if( !ti_sel.IsOk() ) { ti_sel = GetItemParent( ti ) ; }
   }
   /*--( Normally the tree should not be empty )---------------------------*/
   if( ti_sel.IsOk() )
   {  SelectItem( ti_sel ) ; }
   else /*--( Should not happen ... )--------------------------------------*/
   {  change_window_pl( nullptr ) ; }
   /*--( The item data will be deleted too )-------------------------------*/
   Delete( ti ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Remove playlist and clean all data associated to it )----------------*/
void CBookPlaylist::delete_pl( const wxTreeItemId &ti )
{
   /*----------------------------------------------------------------------*/
   CPlaylist *pl = get_playlist( ti ) ;
   if( pl == nullptr || !pl->get_deletable() ) { return ; }
   /*----------------------------------------------------------------------*/
   if( wxMessageBox( wxString::Format( _( "Delete \"%s\" playlist ?" ),
                                       pl->get_label()
                                     ),
                     _( "Confirmation" ), wxYES_NO
                   ) != wxYES
     )
   {  return ; }
   /*--( Clear data associated to this playlist )--------------------------*/
   pl->delete_cfg() ;
   /*--( Page is removed and the window is deleted )-----------------------*/
   remove_pl( ti ) ;
   /*--( Finally delete the playlist )-------------------------------------*/
   delete pl ;
   /*--( Reinitialize parent nodes )---------------------------------------*/
   init_list_parents( GetSelection() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::context_menu( wxTreeEvent &event )
{
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti = event.GetItem() ;
   CPlaylist *pl = get_playlist( ti ) ;
   wxMenu menu ;
   int    i_id = 0 ;
   int    i_id_rename_pl     = -1 ;
   int    i_id_sort_pl       = -1 ;
   int    i_id_add_pl_played = -1 ;

   /*----------------------------------------------------------------------*/
   if( pl->get_renamable() )
   {  i_id_rename_pl = i_id++ ;
      mtf::append_item( &menu, i_id_rename_pl, _( "&Rename\tF2" ),
                        accessories_text_editor_16x16_xpm
                      ) ;
   }
   /*----------------------------------------------------------------------*/
   if( pl->get_order_ch() && pl->get_nb_channel() > 0 )
   {  i_id_sort_pl = i_id++ ;
      mtf::append_item( &menu, i_id_sort_pl, _( "&Sort channels by name" ),
                        preferences_desktop_font_16x16_xpm
                      ) ;
   }
   /*----------------------------------------------------------------------*/
   if( pl->get_deletable() )
   {  mtf::append_item( &menu, wxID_DELETE,
                        wxGetStockLabel( wxID_DELETE, wxSTOCK_WITH_MNEMONIC )
                        + '\t' + _( "Del" ),
                        mtf::get_menu_bmp( wxART_DELETE )
                      ) ;
   }
   /*----------------------------------------------------------------------*/
   if( pl->get_cut_paste_ch() )
   {  /*--( Extract the standard accelerator used )------------------------*/
      const wxString co_s_accel
                   = wxGetStockLabel( wxID_PASTE, wxSTOCK_WITH_ACCELERATOR
                                    ).AfterLast( '\t' ) ;
      /*--( To specify that only channels are pasted, not a playlist )-----*/
      mtf::append_item( &menu, wxID_PASTE,
                        _( "&Paste channels" ) + '\t' + co_s_accel,
                        mtf::get_menu_bmp( wxART_PASTE )
                      ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( pl->get_accept_ch() && !wxGetApp().m_frame->get_vec_media().empty() )
   {  i_id_add_pl_played = i_id++ ;
      mtf::append_item( &menu, i_id_add_pl_played,
                        _( "&Add currently played channels" ),
                        list_add_16x16_xpm
                      ) ;
   }
   /*----------------------------------------------------------------------*/
   if( pl->get_movable() )
   {
      /*--------------------------------------------------------------------+
      ! No key sequences because Alt+Up or Alt+Down key sequence catch is   !
      ! problematic under MSW                                               !
      +--------------------------------------------------------------------*/
      if( GetPrevSibling( ti ).IsOk() )
      {  mtf::append_item( &menu, ID_MOVE_UP, _( "Move up" ),
                           go_up_16x16_xpm
                         ) ;
      }
      /*-------------------------------------------------------------------*/
      if( GetNextSibling( ti ).IsOk() )
      {  mtf::append_item( &menu, ID_MOVE_DOWN, _( "Move down" ),
                           go_down_16x16_xpm
                         ) ;
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( pl->get_nb_channel() > 0 )
   {
      /*-------------------------------------------------------------------*/
      if( menu.GetMenuItemCount() > 0 ) { menu.AppendSeparator() ; }
      /*-------------------------------------------------------------------*/
      mtf::append_item( &menu, ID_PLAY_START_PAUSE,_( "&Play all channels" ),
                        media_playback_start_16x16_xpm
                      ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Useless to go further if nothing can be done )--------------------*/
   if( menu.GetMenuItemCount() == 0 ) { return ; }

   /*--( Check that the chosen tab is active : more user friendly )--------*/
   if( !IsSelected( ti ) ) { SelectItem( ti ) ; }

   /*-----------------------------------------------------------------------+
   ! Display menu and wait for user input                                   !
   +-----------------------------------------------------------------------*/
   int i_sel_id = GetPopupMenuSelectionFromUser( menu ) ;

   /*----------------------------------------------------------------------*/
   if( i_sel_id == i_id_rename_pl )
   {  rename_pl( ti ) ; }
   else
   if( i_sel_id == i_id_sort_pl )
   {  pl->sort_by_name() ; }
   else
   if( i_sel_id == wxID_DELETE )
   {  delete_pl( ti ) ; }
   else
   if( i_sel_id == wxID_PASTE )
   {  pl->paste_from_clipboard() ; }
   else
   if( i_sel_id == i_id_add_pl_played )
   {  pl->add_channel_played() ; }
   if( i_sel_id == ID_MOVE_UP )
   {  move_up( ti ) ; }
   else
   if( i_sel_id == ID_MOVE_DOWN )
   {  move_down( ti ) ; }
   else
   if( i_sel_id == ID_PLAY_START_PAUSE )
   {  wxGetApp().m_frame->play_channel( pl->get_vs_channel() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxTreeItemId CBookPlaylist::find_child( const wxTreeItemId &ti_parent,
                                        const wxString &s_text
                                      )
{
   /*----------------------------------------------------------------------*/
   wxTreeItemIdValue tiv_cookie ;
   wxTreeItemId ti = GetFirstChild( ti_parent, tiv_cookie ) ;
   /*----------------------------------------------------------------------*/
   for( ; ti.IsOk() && GetItemText( ti ) != s_text ;
        ti = GetNextChild( ti_parent, tiv_cookie )
      )
   {  ; }
   /*----------------------------------------------------------------------*/
   return( ti ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxTreeItemId CBookPlaylist::make_path( const wxString &s_path )
{
   /*----------------------------------------------------------------------*/
   wxTreeItemId ti = GetRootItem() ;
   /*----------------------------------------------------------------------*/
   wxFileName fn_path ;
   fn_path.AssignDir( s_path ) ;
   wxFileName fn_current ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_dir : fn_path.GetDirs() )
   {
      /*--( If this element doesn't exist it is created )------------------*/
      wxTreeItemId ti_dir = find_child( ti, s_dir ) ;
      if( ti_dir.IsOk() )
      {  ti = ti_dir ; }
      else
      {  CPlaylist *pl = new CPlaylistNode( wxGetApp().m_frame, s_dir, s_dir,
                                            fn_current.GetFullPath()
                                          ) ;
         ti = add_pl( pl ) ;
      }
      /*-------------------------------------------------------------------*/
      fn_current.AppendDir( s_dir )  ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   return( ti ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Recursive function )-------------------------------------------------*/
void CBookPlaylist::exec_on_tree_pl( t_fnc_pl fnc_pl,
                                     wxTreeItemId ti_start
                                   )
{
   /*----------------------------------------------------------------------*/
   if( !ti_start.IsOk() ) { return ; }
   /*--( Exec on the subtree )---------------------------------------------*/
   if( ItemHasChildren( ti_start ) )
   {
      /*-------------------------------------------------------------------*/
      wxTreeItemIdValue tiv_cookie ;
      /*-------------------------------------------------------------------*/
      for( auto ti_child = GetFirstChild( ti_start, tiv_cookie ) ;
           ti_child.IsOk() ;
           ti_child = GetNextChild( ti_start, tiv_cookie )
         )
      {  exec_on_tree_pl( fnc_pl, ti_child ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*--( Exec on current element )-----------------------------------------*/
   CPlaylist *pl_start = get_playlist( ti_start ) ;
   if( pl_start != nullptr ) { fnc_pl( pl_start ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CBookPlaylist::init_list_parents( const wxTreeItemId &ti_org )
{
   /*----------------------------------------------------------------------*/
   for( wxTreeItemId ti = GetItemParent( ti_org ) ;
        ti.IsOk() && ti != GetRootItem() ;
        ti = GetItemParent( ti )
      )
   {  /*-------------------------------------------------------------------*/
      CPlaylist *pl = get_playlist( ti ) ;
      if( pl != nullptr ) { pl->init_list() ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                      End of file CBookPlaylist.cpp                        !
+==========================================================================*/
