/*====================================+=====================================+
! File CPlaylist.cpp                  ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/imaglist.h>
#include <wx/srchctrl.h>
#include <wx/tokenzr.h>
#include <wx/clipbrd.h>
#include <wx/statline.h>
/*-------------------------------------------------------------------------*/
#include "matelefree.h"
#include "CApp.h"
#include "CFrame.h"
#include "CPlaylist.h"
#include "CDndChannel.h"
#include "CPlSubWin.h"
#include "CDlgChName.h"
#include "tango/applications_other_24x24.xpm"
#include "tango/accessories_text_editor_16x16.xpm"
#include "tango/document_new_16x16.xpm"
#include "tango/media_playback_start_16x16.xpm"
#include "tango/go_up_16x16.xpm"
#include "tango/go_down_16x16.xpm"
#include "tango/list_add_16x16.xpm"
#include "tango/start_here_16x16.xpm"
#include "tango/emblem_important_16x16.xpm"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
int CPlaylist::m_st_i_col_width = -1 ;
/*-------------------------------------------------------------------------*/



/*--( The column will be created outside of the constructor )--------------*/
CPlaylist::CPlaylist( wxWindow *parent, const wxString &s_name,
                      const wxString &s_label,
                      const wxString &s_tree_path
                    )
         : wxListCtrl( parent, wxID_ANY, wxDefaultPosition, wxDefaultSize,
                       wxLC_REPORT | wxLC_NO_HEADER | wxLC_VIRTUAL
                     ),
           m_s_tree_path( s_tree_path ),
           m_s_name( s_name )
{
   /*----------------------------------------------------------------------*/
   SetName( m_s_name ) ;
   SetDropTarget( new CChannelDropTargetOnPlaylist( this ) ) ;
   /*--( By default the name is used as label )----------------------------*/
   m_s_label = ( s_label.empty() ? m_s_name : s_label ) ;

   /*--( Col width is computed outside )-----------------------------------*/
   AppendColumn( wxEmptyString, wxLIST_FORMAT_LEFT, m_st_i_col_width ) ;

   /*----------------------------------------------------------------------*/
   Bind( EVT_AST_CONTEXT_MENU,
         [ this ]( wxCommandEvent &event )
         {  add_streams_to_context_menu( event.GetString() ) ; }
       ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_LIST_ITEM_ACTIVATED,
         [ this ]( wxListEvent &event )
         {  play_item( event.GetIndex() ) ; }
       ) ;
   Bind( wxEVT_LIST_ITEM_RIGHT_CLICK,
         [ this ]( wxListEvent &event ) { context_menu( event ) ; }
       ) ;
   Bind( wxEVT_LIST_BEGIN_DRAG, &CPlaylist::OnBeginDrag, this ) ;
   Bind( wxEVT_KEY_DOWN, &CPlaylist::OnKeyDown, this ) ;

   /*--( The list will be deleted in the destructor )----------------------*/
   SetImageList( wxGetApp().m_il_logo, wxIMAGE_LIST_SMALL ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::OnGetItemImage( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_channel == nullptr ) { return( -1 ) ; }
   /*----------------------------------------------------------------------*/
   return( p_channel->get_logo_il_pos() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CPlaylist::OnGetItemText( long l_item, long l_column ) const
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_channel == nullptr || l_column != 0 )
   {  return( wxEmptyString ) ; }
   /*--( Outside of the source, the source name is added )-----------------*/
   if( get_text_long() )
   {  return( p_channel->get_pl_text_long() ) ; }
   /*----------------------------------------------------------------------*/
   return( p_channel->get_pl_text_short() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxListItemAttr *CPlaylist::OnGetItemAttr( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   CApp::e_lia     lia     = CApp::LIA_NORMAL ;
   CApp::e_lia_air lia_air = CApp::LIAAIR_OFF ;
   CApp::e_lia_alt lia_alt
                 = ( l_item % 2 == 0 ? CApp::LIAALT_OFF : CApp::LIAALT_ON ) ;

   /*--( Modify attribute based on the channel state )---------------------*/
   CChannel *p_channel = get_channel( l_item ) ;
   if( p_channel != nullptr )
   {
      /*-------------------------------------------------------------------*/
      if(    get_use_attr_onair()
          && p_channel->get_use_attr_onair()
          && p_channel->get_onair()
        )
      {  lia_air = CApp::LIAAIR_ON ; }

      /*-------------------------------------------------------------------*/
      CFrame *frame = wxGetApp().m_frame ;
      const CVlcMedia *p_media_channel = frame->find_media( p_channel ) ;
      if( p_media_channel != nullptr )
      {  lia = ( p_media_channel == frame->get_media_active()
                 ? CApp::LIA_PLAYED_ACTIVE : CApp::LIA_PLAYED
               ) ;
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   return( wxGetApp().get_lia( lia, lia_air, lia_alt ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::OnBeginDrag( wxListEvent &event )
{
   /*----------------------------------------------------------------------*/
   int i_flags = ( get_give_ch() ? wxDrag_AllowMove : wxDrag_CopyOnly ) ;
   CChannelDataObject data_oc( this ) ;
   /*----------------------------------------------------------------------*/
   set_drag_in_progress( true ) ;
   wxDropSource drop_source( data_oc, this ) ;
   wxDragResult drag_result = drop_source.DoDragDrop( i_flags ) ;
   /*--( Move out of this pl ? if so we have to remove the source )--------*/
   if( drag_result == wxDragMove && !get_save_cfg() )
   {  del_channel_hil() ; }
   /*----------------------------------------------------------------------*/
   set_drag_in_progress( false ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::OnKeyDown( wxKeyEvent &event )
{
   /*----------------------------------------------------------------------*/
   bool boo_cmd_key
      = ( ( event.GetModifiers() & ~wxMOD_ALT ) == wxMOD_CMD ) ;
   bool boo_alt_key
      = ( ( event.GetModifiers() & wxMOD_ALT ) == wxMOD_ALT ) ;
   /*----------------------------------------------------------------------*/
   if( boo_alt_key && event.GetKeyCode() == WXK_RETURN )
   {  play_item( get_item_focus(), true ) ; }
   else
   if( boo_cmd_key && event.GetKeyCode() == 'A' )
   {  set_item_hil( -1, true ) ; }
   else
   if( boo_alt_key && event.GetKeyCode() == WXK_UP )
   {  move_channel_hil( -1 ) ; }
   else
   if( boo_alt_key && event.GetKeyCode() == WXK_DOWN )
   {  move_channel_hil( +1 ) ; }
   else
   if( event.GetKeyCode() == WXK_F2 )
   {  ren_channel_hil() ; }
   else
   if( event.GetKeyCode() == WXK_DELETE )
   {  del_channel_hil() ; }
   else
   if( boo_cmd_key && event.GetKeyCode() == 'C' )
   {  copy_hil_to_clipboard() ; }
   else
   if( boo_cmd_key && event.GetKeyCode() == 'X' )
   {  cut_hil_to_clipboard() ; }
   else
   if( boo_cmd_key && event.GetKeyCode() == 'V' )
   {  paste_from_clipboard() ; }
   else
   {  event.Skip() ; }
   /*----------------------------------------------------------------------*/
}

/*--( Image of the playlist in the playlist tree )-------------------------*/
void CPlaylist::init_image()
{
   /*----------------------------------------------------------------------*/
   const mtf::t_map_strstr *p_map_image = &wxGetApp().m_map_image_fn ;
   /*--( Try to find an icon named after the label or the name )-----------*/
   mtf::t_map_strstr_cit cit_img
                  = p_map_image->find( mtf::prepare_string( get_label() ) ) ;
   if( cit_img == p_map_image->end() && get_name() != get_label() )
   {  cit_img = p_map_image->find( mtf::prepare_string( get_name() ) ) ; }

   /*----------------------------------------------------------------------*/
   CBookPlaylist *bopl = wxGetApp().m_frame->get_bopl() ;
   wxBitmap      bmp ;
   /*----------------------------------------------------------------------*/
   if( cit_img != p_map_image->end() )
   {  /*-------------------------------------------------------------------*/
      wxLogNull no_log ; // Suppress useless warnings
      bmp.LoadFile( cit_img->second, wxBITMAP_TYPE_PNG ) ;
      /*-------------------------------------------------------------------*/
   }
   /*--( No file found or the found one is invalid )-----------------------*/
   if( !bmp.IsOk() || bmp.GetSize() != bopl->get_il_pl()->GetSize() )
   {  bmp = wxBitmap( applications_other_24x24_xpm ) ; }
   /*--( Add or replace current image )------------------------------------*/
   if( m_i_il_pos < 0 )
   {  m_i_il_pos = bopl->get_il_pl()->Add( bmp ) ; }
   else /*--( Replace current one )----------------------------------------*/
   {  bopl->get_il_pl()->Replace( m_i_il_pos, bmp ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::get_image_il_pos()
{
   /*----------------------------------------------------------------------*/
   if( m_i_il_pos < 0 ) { init_image() ; }
   /*----------------------------------------------------------------------*/
   return( m_i_il_pos ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylist *CPlaylist::st_new_pl_source( const wxString &s_name,
                                        const wxString &s_label,
                                        const wxString &s_tree_path,
                                        mtf::t_vec_string &vs_channel
                                      )
{
   /*----------------------------------------------------------------------*/
   CPlaylist *pl = new CPlaylist( wxGetApp().m_frame, s_name,
                                  wxGetTranslation( s_label ), s_tree_path
                                ) ;
   /*----------------------------------------------------------------------*/
   pl->set_text_long( false ) ;
   pl->swap_vs_channel( vs_channel ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylist *CPlaylist::st_new_pl_source( e_pl_known pl_known,
                                        const wxString &s_tree_path,
                                        mtf::t_vec_string &vs_channel
                                      )
{
   /*----------------------------------------------------------------------*/
   CPlaylist *pl
   = st_new_pl_source( wxGetApp().get_pk_name( pl_known ),
                       wxGetTranslation(wxGetApp().get_pk_label( pl_known )),
                       s_tree_path, vs_channel
                     ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( pl_known, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistFilter *CPlaylist::st_new_pl_fbtvb()
{
   /*----------------------------------------------------------------------*/
   CPlaylistFilter *pl
   = new CPlaylistFilter( wxGetApp().m_frame,
                     wxGetApp().get_pk_name( PK_FBTVB ),
                     wxGetTranslation( wxGetApp().get_pk_label( PK_FBTVB ) ),
                          wxGetApp().m_s_tree_fbtv_path,
                          wxGetApp().get_pk_pl( PK_FBTV ),
                          []( const CChannel *p_channel ) -> bool
                          {  return(     p_channel->get_pl_known() == PK_FBTV
                                     && !p_channel->is_fbtvo()
                                   ) ;
                          }
                        ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( PK_FBTVB, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistFilter *CPlaylist::st_new_pl_fbtvo()
{
   /*----------------------------------------------------------------------*/
   CPlaylistFilter *pl
   = new CPlaylistFilter( wxGetApp().m_frame,
                     wxGetApp().get_pk_name( PK_FBTVO ),
                     wxGetTranslation( wxGetApp().get_pk_label( PK_FBTVO ) ),
                          wxGetApp().m_s_tree_fbtv_path,
                          wxGetApp().get_pk_pl( PK_FBTV ),
                          []( const CChannel *p_channel ) -> bool
                          {  return( p_channel->is_fbtvo() ) ; }
                        ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( PK_FBTVO, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistFbtvoOk *CPlaylist::st_new_pl_fbtvo_ok(
                                               const wxString &s_channel_list
                                               )
{
   /*----------------------------------------------------------------------*/
   CPlaylistFbtvoOk *pl
   = new CPlaylistFbtvoOk( wxGetApp().m_frame,
                  wxGetApp().get_pk_name( PK_FBTVO_OK ),
                  wxGetTranslation( wxGetApp().get_pk_label( PK_FBTVO_OK ) ),
                           wxGetApp().m_s_tree_fbtv_path
                          ) ;
   /*----------------------------------------------------------------------*/
   pl->string_to_datas( s_channel_list ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( PK_FBTVO_OK, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistNode *CPlaylist::st_new_pl_live_node()
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_name = wxGetApp().get_pk_name( PK_LIVE_NODE ) ;
   const wxString co_s_label
              = wxGetTranslation( wxGetApp().get_pk_label( PK_LIVE_NODE ) ) ;
   wxGetApp().m_s_tree_live_path = co_s_label ;
   /*----------------------------------------------------------------------*/
   CPlaylistLiveNode *pl
       = new CPlaylistLiveNode( wxGetApp().m_frame, co_s_name, co_s_label,
                                wxEmptyString
                              ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( PK_LIVE_NODE, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistLive *CPlaylist::st_new_pl_live( e_pl_known pl_known,
                                          wxConfigBase *p_conf
                                        )
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   const wxString co_s_name = wxGetApp().get_pk_name( pl_known ) ;
   const wxString co_s_label
                = wxGetTranslation( wxGetApp().get_pk_label( pl_known ) ) ;
   /*----------------------------------------------------------------------*/
   CPlaylistLive *pl ;
   switch( pl_known )
   {  /*-------------------------------------------------------------------*/
      case PK_LIVE_DM :
         pl = new CPlaylistDailymotion( frame, co_s_name, co_s_label,
                                        wxGetApp().m_s_tree_live_path
                                      ) ;
         break ;
      /*-------------------------------------------------------------------*/
      case PK_LIVE_TW :
         pl = new CPlaylistTwitch( frame, co_s_name, co_s_label,
                                   wxGetApp().m_s_tree_live_path
                                 ) ;
         break ;
      /*-------------------------------------------------------------------*/
      case PK_LIVE_YT :
         pl = new CPlaylistYoutube( frame, co_s_name, co_s_label,
                                    wxGetApp().m_s_tree_live_path
                                  ) ;
         break ;
      /*-------------------------------------------------------------------*/
      case PK_LIVE_US :
         pl = new CPlaylistUstream( frame, co_s_name, co_s_label,
                                    wxGetApp().m_s_tree_live_path
                                  ) ;
         break ;
      /*-------------------------------------------------------------------*/
      case PK_LIVE_PE :
         pl = new CPlaylistPeriscope( frame, co_s_name, co_s_label,
                                      wxGetApp().m_s_tree_live_path
                                    ) ;
         break ;
      /*-------------------------------------------------------------------*/
      case PK_LIVE_MX :
         pl = new CPlaylistMixer( frame, co_s_name, co_s_label,
                                  wxGetApp().m_s_tree_live_path
                                ) ;
         break ;
      /*-------------------------------------------------------------------*/
      default : return( nullptr ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Create channels and add them to the playlist )--------------------*/
   wxString s_channel_list ;
   if( p_conf->Read( co_s_name, &s_channel_list, "" ) )
   {  pl->string_to_datas( s_channel_list ) ; }
   pl->init_access_onair() ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( pl_known, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistUserNode *CPlaylist::st_new_pl_user_node()
{
   /*----------------------------------------------------------------------*/
   wxString s_label
          = wxGetTranslation( wxGetApp().get_pk_label( PK_USER_NODE ) ) ;
   wxGetApp().m_s_tree_pl_user_path = s_label ;
   /*----------------------------------------------------------------------*/
   CPlaylistUserNode *pl
   = new CPlaylistUserNode( wxGetApp().m_frame,
                            wxGetApp().get_pk_name( PK_USER_NODE ),
                            s_label, ""
                          ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( PK_USER_NODE, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistUser *CPlaylist::st_new_pl_user( const wxString &s_text,
                                          const wxString &s_channel_list
                                        )
{
   /*----------------------------------------------------------------------*/
   CPlaylistUser *pl
   = new CPlaylistUser( wxGetApp().m_frame, s_text, s_text,
                        wxGetApp().m_s_tree_pl_user_path
                      ) ;
   /*----------------------------------------------------------------------*/
   pl->string_to_datas( s_channel_list ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistSearch *CPlaylist::st_new_pl_search()
{
   /*----------------------------------------------------------------------*/
   CPlaylistSearch *pl
   = new CPlaylistSearch( wxGetApp().m_frame,
                    wxGetApp().get_pk_name( PK_SEARCH ),
                    wxGetTranslation( wxGetApp().get_pk_label( PK_SEARCH ) ),
                          ""
                        ) ;
   /*----------------------------------------------------------------------*/
   wxGetApp().set_pk_pl( PK_SEARCH, pl ) ;
   /*----------------------------------------------------------------------*/
   return( pl ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::init_item_count()
{
   /*----------------------------------------------------------------------*/
   long l_item_focus = get_item_focus() ;
   /*----------------------------------------------------------------------*/
   SetItemCount( get_nb_channel() ) ;
   /*----------------------------------------------------------------------*/
   if( !m_vs_channel.empty() )
   {  /*-------------------------------------------------------------------*/
      if( l_item_focus < 0 )
      {  set_item_active( 0 ) ; }
      else
      if( l_item_focus >= get_nb_channel() )
      {  set_item_active( get_nb_channel() - 1 ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   Refresh() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::string_to_datas( const wxString &s_val )
{
   /*----------------------------------------------------------------------*/
   wxStringTokenizer tkz( s_val, '\n' ) ;
   wxString s_token ;
   /*----------------------------------------------------------------------*/
   while( tkz.HasMoreTokens() )
   {
      /*-------------------------------------------------------------------*/
      s_token = tkz.GetNextToken() ;
      /*--------------------------------------------------------------------+
      ! Invalid channel names are kept in case of "come back".              !
      ! For example if the freebox access is temporarily removed or lost.   !
      +--------------------------------------------------------------------*/
      if( wxGetApp().get_channel( s_token ) != nullptr )
      {  m_vs_channel.emplace_back( s_token ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CPlaylist::datas_to_string() const
{
   /*----------------------------------------------------------------------*/
   wxString s_res ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : m_vs_channel )
   {  /*-------------------------------------------------------------------*/
      if( !s_res.empty() ) { s_res += '\n' ; }
      /*-------------------------------------------------------------------*/
      s_res += s_channel ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( s_res ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::sort_by_name()
{
   /*----------------------------------------------------------------------*/
   if( !get_order_ch() ) { return ; }
   /*--( The focus is kept )-----------------------------------------------*/
   int i_focus = get_item_focus() ;
   /*----------------------------------------------------------------------*/
   std::sort( m_vs_channel.begin(), m_vs_channel.end(),
              []( const wxString &s1, const wxString &s2 )
              {  return( s1.CmpNoCase( s2 ) < 0 ) ; }
            ) ;
   /*--( Focus set back )--------------------------------------------------*/
   set_item_focus( i_focus ) ;
   /*--( Reinitialize parent nodes )---------------------------------------*/
   set_save_cfg( true ) ;
   /*--( Refresh the playlist : items have move around )-------------------*/
   Refresh() ;
   /*----------------------------------------------------------------------*/
   wxGetApp().m_frame->get_bopl()->init_list_parents( this ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CPlaylist::copy_hil_to_clipboard()
{
   /*----------------------------------------------------------------------*/
   if( !wxTheClipboard->Open() ) { return( false ) ; }
   /*----------------------------------------------------------------------*/
   wxTheClipboard->AddData( new CChannelDataObject( this ) ) ;
   wxTheClipboard->Close() ;
   /*----------------------------------------------------------------------*/
   return( true ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CPlaylist::cut_hil_to_clipboard()
{
   /*----------------------------------------------------------------------*/
   if( !get_give_ch() ) { return( false ) ; }
   /*----------------------------------------------------------------------*/
   bool boo_ok = copy_hil_to_clipboard() ;
   if( boo_ok ) { del_channel_hil() ; }
   /*----------------------------------------------------------------------*/
   return( boo_ok ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CPlaylist::paste_from_clipboard()
{
   /*----------------------------------------------------------------------*/
   if( !get_accept_ch() ) { return( false ) ; }
   /*----------------------------------------------------------------------*/
   if( !wxTheClipboard->Open() ) { return( false ) ; }
   /*----------------------------------------------------------------------*/
   CChannelDataObject cdo ;
   /*----------------------------------------------------------------------*/
   if( wxTheClipboard->GetData( cdo ) )
   {  add_channel( cdo.data_to_vs(), get_item_focus() ) ; }
   /*----------------------------------------------------------------------*/
   wxTheClipboard->Close() ;
   /*----------------------------------------------------------------------*/
   return( true ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::add_channel( const mtf::t_vec_string &vs_channel,
                            int i_ins_pos
                          )
{
   /*----------------------------------------------------------------------*/
   if( vs_channel.empty() ) { return( 0 ) ; }
   /*----------------------------------------------------------------------*/
   int i_nb_added = 0 ;
   /*--( By default an append is done )------------------------------------*/
   if( i_ins_pos < 0 ) { i_ins_pos = m_vs_channel.size() ; }
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : vs_channel )
   {  /*--( Channel is added if it exists and is not already present )-----*/
      if(    get_channel_item( s_channel ) < 0
          && wxGetApp().get_channel( s_channel ) != nullptr
        )
      {  m_vs_channel.emplace( m_vs_channel.begin() + i_ins_pos, s_channel );
         ++i_ins_pos ;
         ++i_nb_added ;
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( i_nb_added > 0 )
   {  /*-------------------------------------------------------------------*/
      init_list() ;
      /*-------------------------------------------------------------------*/
      if( get_nb_channel() == 1 ) { set_item_active( 0 ) ; }
      /*-------------------------------------------------------------------*/
      set_save_cfg( true ) ;
      /*-------------------------------------------------------------------*/
      wxGetApp().m_frame->get_bopl()->init_list_parents( this ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( i_nb_added ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::add_channel_hil( CPlaylist *pl_src, int i_ins_pos )
{
   /*----------------------------------------------------------------------*/
   if( !get_accept_ch() || pl_src->get_nb_hil() == 0 ) { return( 0 ) ; }
   /*----------------------------------------------------------------------*/
   mtf::t_vec_string vs_channel ;
   pl_src->get_hil( vs_channel ) ;
   /*----------------------------------------------------------------------*/
   return( add_channel( vs_channel, i_ins_pos ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::add_channel_played()
{
   /*----------------------------------------------------------------------*/
   mtf::t_vec_string vs_channel ;
   /*----------------------------------------------------------------------*/
   for( const auto p_media : wxGetApp().m_frame->get_vec_media() )
   {  vs_channel.emplace_back( p_media->get_channel_name() ) ; }
   /*----------------------------------------------------------------------*/
   return( add_channel( vs_channel ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::del_channel( const mtf::t_vec_string &vs_channel )
{
   /*----------------------------------------------------------------------*/
   if( !get_delete_ch() || vs_channel.empty() ) { return( 0 ) ; }
   /*----------------------------------------------------------------------*/
   int i_nb_deleted = 0 ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : vs_channel )
   {  /*-------------------------------------------------------------------*/
      long l_item = get_channel_item( s_channel ) ;
      /*-------------------------------------------------------------------*/
      if( l_item >= 0 )
      {  m_vs_channel.erase( m_vs_channel.begin() + l_item ) ;
         ++i_nb_deleted ;
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( i_nb_deleted > 0 )
   {  /*-------------------------------------------------------------------*/
      init_list() ;
      set_save_cfg( true ) ;
      wxGetApp().m_frame->get_bopl()->init_list_parents( this ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( i_nb_deleted ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::del_channel_hil()
{
   /*----------------------------------------------------------------------*/
   if( !get_delete_ch() || get_nb_hil() == 0 ) { return( 0 ) ; }
   /*----------------------------------------------------------------------*/
   mtf::t_vec_string vs_channel ;
   get_hil( vs_channel ) ;
   /*--( The channels will disappear : keep only one selected )------------*/
   set_item_active( get_item_focus() ) ;
   /*----------------------------------------------------------------------*/
   return( del_channel( vs_channel ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::ren_channel_hil()
{
   /*----------------------------------------------------------------------*/
   if( !get_rename_ch() || get_nb_hil() != 1 ) { return ; }
   /*----------------------------------------------------------------------*/
   long l_item = get_first_item_hil() ;
   CChannel *p_channel = get_channel( l_item ) ;
   if( p_channel == nullptr ) { return ; }

   /*----------------------------------------------------------------------*/
   wxString s_from_full = p_channel->get_name() ;
   wxString s_to_short ;
   /*--( User enters the new name )----------------------------------------*/
   CDlgChName( nullptr, _( "Rename Channel" ), p_channel
             ).get_modal( s_to_short ) ;
   /*--( The user can cancel the renaming )--------------------------------*/
   if( s_to_short.empty() ) { return ; }

   /*----------------------------------------------------------------------*/
   wxGetApp().rename_channel( s_from_full, s_to_short ) ;

   /*--( Rename its references in the user playlists )---------------------*/
   t_vec_pl vec_pl_user ;
   CBookPlaylist *bopl = wxGetApp().m_frame->get_bopl() ;
   bopl->get_pl_children( PK_USER_NODE, vec_pl_user ) ;
   /*----------------------------------------------------------------------*/
   for( auto pl : vec_pl_user )
   {
      /*-------------------------------------------------------------------*/
      CPlaylistUser *pl_user = static_cast< CPlaylistUser * >( pl ) ;
      /*--( A channel is supposed to be unique in a playlist )-------------*/
      pl_user->change_channel_name( s_from_full, p_channel->get_name() ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( !vec_pl_user.empty() )
   {  wxGetApp().get_pk_pl( PK_USER_NODE )->init_list() ; }

   /*--( "Search" can list this channel too )------------------------------*/
   wxGetApp().get_pk_pl( PK_SEARCH )->change_channel_name(
                                           s_from_full, p_channel->get_name()
                                                         ) ;
   /*--( Rename the channel in this list )---------------------------------*/
   m_vs_channel[ l_item ] = p_channel->get_name() ;
   bopl->init_list_parents( this ) ;
   /*----------------------------------------------------------------------*/
   set_save_cfg( true ) ;
   RefreshItem( l_item ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::change_channel_name( const wxString &s_from,
                                     const wxString &s_to
                                   )
{
   /*----------------------------------------------------------------------*/
   long l_item = get_channel_item( s_from ) ;
   if( l_item >= 0 )
   {  /*-------------------------------------------------------------------*/
      m_vs_channel[ l_item ] = s_to ;
      set_save_cfg( true ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::swap_channel( int i_item1, int i_item2 )
{
   /*----------------------------------------------------------------------*/
   if( !get_order_ch() ) { return ; }
   /*----------------------------------------------------------------------*/
   const long co_state = wxLIST_STATE_FOCUSED | wxLIST_STATE_SELECTED ;
   long  l_state1 = GetItemState( i_item1, co_state ) ;
   long  l_state2 = GetItemState( i_item2, co_state ) ;
   /*----------------------------------------------------------------------*/
   std::swap( m_vs_channel[ i_item1 ], m_vs_channel[ i_item2 ] ) ;
   /*----------------------------------------------------------------------*/
   SetItemState( i_item1, l_state2, co_state ) ;
   SetItemState( i_item2, l_state1, co_state ) ;
   /*----------------------------------------------------------------------*/
   RefreshItem( i_item1 ) ;
   RefreshItem( i_item2 ) ;
   /*----------------------------------------------------------------------*/
}

/*--( The number of moves can be negative (up) or positive (down) )--------*/
void CPlaylist::move_channel_hil( const int co_i_nb_move )
{
   /*----------------------------------------------------------------------*/
   if( !get_order_ch() || co_i_nb_move == 0 ) { return ; }
   /*----------------------------------------------------------------------*/
   mtf::t_vec_int vi_hil ;
   if( get_hil( vi_hil ) == 0 ) { return ; }

   /*--( Don't move after the edges )--------------------------------------*/
   if(    vi_hil[ 0 ] + co_i_nb_move < 0
       || vi_hil.back() + co_i_nb_move >= GetItemCount()
     )
   {  return ; }
   /*----------------------------------------------------------------------*/
   const int co_i_nb_hil = static_cast< int >( vi_hil.size() ) ;

   /*--( Move step by step )-----------------------------------------------*/
   if( co_i_nb_move < 0 )
   {  /*-------------------------------------------------------------------*/
      for( int i_num_move = 0 ; i_num_move > co_i_nb_move ; --i_num_move )
      {  /*----------------------------------------------------------------*/
         for( int i_hil = 0 ; i_hil < co_i_nb_hil ; ++i_hil )
         {  swap_channel( vi_hil[ i_hil ], vi_hil[ i_hil ] - 1 ) ;
            --vi_hil[ i_hil ] ;
         }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   else
   {  /*-------------------------------------------------------------------*/
      for( int i_num_move = 0 ; i_num_move < co_i_nb_move ; ++i_num_move )
      {  /*----------------------------------------------------------------*/
         for( int i_hil = co_i_nb_hil - 1 ; i_hil >= 0 ; --i_hil )
         {  swap_channel( vi_hil[ i_hil ], vi_hil[ i_hil ] + 1 ) ;
            ++vi_hil[ i_hil ] ;
         }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   set_item_visible( vi_hil[ 0 ] ) ;
   /*----------------------------------------------------------------------*/
   SetFocus() ;
   set_save_cfg( true ) ;
   wxGetApp().m_frame->get_bopl()->init_list_parents( this ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::append_channel_from( const CPlaylist *pl )
{
   /*----------------------------------------------------------------------*/
   if( pl == nullptr ) { return ; }
   /*-----------------------------------------------------------------------+
   ! Normally (except user node) the channels are unique to each playlist.  !
   ! This is why we don't look for duplicates.                              !
   +-----------------------------------------------------------------------*/
   m_vs_channel.insert( m_vs_channel.end(),
                        pl->m_vs_channel.begin(), pl->m_vs_channel.end()
                      ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::rename( const wxString &s_name )
{
   /*----------------------------------------------------------------------*/
   if( !get_renamable() ) { return ; }
   /*--( First rename the ".ini" entry (if any) )--------------------------*/
   wxConfigBase *p_conf = wxConfigBase::Get() ;
   p_conf->SetPath( CPlaylist::m_st_co_p_c_cfg_path ) ;
   p_conf->RenameEntry( m_s_name, s_name ) ;
   /*--( Then change the name itself and reload the image (based on name) )*/
   m_s_name = m_s_label = s_name ;
   SetName( m_s_name ) ;
   init_image() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::delete_cfg()
{
   /*----------------------------------------------------------------------*/
   wxConfigBase *p_conf = wxConfigBase::Get() ;
   p_conf->SetPath( CPlaylist::m_st_co_p_c_cfg_path ) ;
   p_conf->DeleteEntry( get_name() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CChannel *CPlaylist::get_channel( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   return( l_item < 0 || l_item >= get_nb_channel()
           ? nullptr
           : wxGetApp().get_channel( m_vs_channel[ l_item ] )
         ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
long CPlaylist::get_channel_item( const wxString &s_channel ) const
{
   /*----------------------------------------------------------------------*/
   auto it = std::find( m_vs_channel.begin(), m_vs_channel.end(), s_channel);
   /*----------------------------------------------------------------------*/
   return( it != m_vs_channel.end() ? it - m_vs_channel.begin() : -1 ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Maybe for this one is there a better way ... )-----------------------*/
long CPlaylist::get_last_item_hil() const
{
   /*----------------------------------------------------------------------*/
   for( long l_item = GetItemCount() - 1 ; l_item >= 0 ; --l_item )
   {  if( is_item_hil( l_item ) ) { return( l_item ) ; } }
   /*----------------------------------------------------------------------*/
   return( -1 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::set_item_hil( const mtf::t_vec_string &vs_channel,
                              bool boo_on
                            )
{
   /*----------------------------------------------------------------------*/
   for( const wxString &s_channel : vs_channel )
   {
      /*-------------------------------------------------------------------*/
      long l_item = get_channel_item( s_channel ) ;
      if( l_item >= 0 ) { set_item_hil( l_item, boo_on ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::get_hil( mtf::t_vec_string &vs_channel ) const
{
   /*----------------------------------------------------------------------*/
   for( long l_item = -1 ; ( l_item = get_next_item_hil( l_item ) ) != -1 ; )
   {  vs_channel.emplace_back( m_vs_channel[ l_item ] ) ; }
   /*----------------------------------------------------------------------*/
   return( vs_channel.size() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::get_hil( mtf::t_vec_int &vi_item ) const
{
   /*----------------------------------------------------------------------*/
   for( long l_item = -1 ; ( l_item = get_next_item_hil( l_item ) ) != -1 ; )
   {  vi_item.emplace_back( l_item ) ; }
   /*----------------------------------------------------------------------*/
   return( vi_item.size() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::set_item_active( long l_item )
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= GetItemCount() ) { return ; }
   /*----------------------------------------------------------------------*/
   set_item_hil( -1, false ) ;
   set_item_state( l_item, wxLIST_STATE_FOCUSED | wxLIST_STATE_SELECTED ) ;
   set_item_visible( l_item ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::refresh_item( long l_item )
{
   /*----------------------------------------------------------------------*/
   if( l_item >= 0 && l_item < static_cast< long >( m_vs_channel.size() ) )
   {  RefreshItem( l_item ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::refresh_item( const wxString &s_channel )
{
   /*----------------------------------------------------------------------*/
   RefreshItem( get_channel_item( s_channel ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::play_item( long l_item, bool boo_new_media )
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= get_nb_channel() ) { return ; }
   /*----------------------------------------------------------------------*/
   set_item_visible( l_item ) ;
   /*-----------------------------------------------------------------------+
   ! Necessary to be sure to have the logo loaded because if an error       !
   ! occurs the display of the message won't show it ...                    !
   +-----------------------------------------------------------------------*/
   Update() ;
   /*--( A channel is always "started" in the main media window )----------*/
   wxGetApp().m_frame->play_channel( m_vs_channel[ l_item ], boo_new_media );
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::play_focus()
{
   /*----------------------------------------------------------------------*/
   long l_item = get_item_focus() ;
   /*----------------------------------------------------------------------*/
   if( l_item >= 0 ) { play_item( l_item ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::play_prev_hil()
{
   /*----------------------------------------------------------------------*/
   long l_item_focus = get_item_focus() ;
   /*--( The "starting point" is the focused item )------------------------*/
   if( l_item_focus < 0 ) { return ; }

   /*----------------------------------------------------------------------*/
   long l_item ;
   /*--( Finding the previous item to play )-------------------------------*/
   if( get_nb_hil() > 1 )
   {  /*--( More than one hil ? the prev hil in the list is played )-------*/
      l_item = get_prev_item_hil( l_item_focus ) ;
      if( l_item < 0 ) { l_item = get_prev_item_hil( -1 ) ; }
      /*-------------------------------------------------------------------*/
   }
   else
   {  /*--( Only one hil ? the prev channel in the list is played )--------*/
      l_item = l_item_focus - 1 ;
      if( l_item < 0 ) { l_item = get_nb_item() - 1 ; }
      set_item_hil( l_item_focus, false ) ;
      set_item_hil( l_item, true ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   set_item_focus( l_item ) ;
   play_item( l_item ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::play_next_hil()
{
   /*----------------------------------------------------------------------*/
   long l_item_focus = get_item_focus() ;
   /*--( The "starting point" is the focused item )------------------------*/
   if( l_item_focus < 0 ) { return ; }

   /*----------------------------------------------------------------------*/
   long l_item ;
   /*--( Finding the next item to play )-----------------------------------*/
   if( get_nb_hil() > 1 )
   {  /*--( More than one hil ? the next hil in the list is played )-------*/
      l_item = get_next_item_hil( l_item_focus ) ;
      if( l_item < 0 ) { l_item = get_next_item_hil( -1 ) ; }
      /*-------------------------------------------------------------------*/
   }
   else
   {  /*--( Only one hil ? the prev channel in the list is played )--------*/
      l_item = l_item_focus + 1 ;
      if( l_item >= get_nb_item() ) { l_item = 0 ; }
      set_item_hil( l_item_focus, false ) ;
      set_item_hil( l_item, true ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   set_item_focus( l_item ) ;
   play_item( l_item ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenu *CPlaylist::build_track_menu( CVlcMedia *p_media,
                                     libvlc_track_type_t type,
                                     int i_id_start
                                   ) const
{
   /*----------------------------------------------------------------------*/
   wxMenu *menu = new wxMenu ;
   int    i_active_track ;
   libvlc_track_description_t *p_tdl ;
   bool   boo_add_deactivate = true ;

   /*--( Item to check in the menu )---------------------------------------*/
   switch( type )
   {
      /*-------------------------------------------------------------------*/
      case libvlc_track_video :
         i_active_track = p_media->get_video_track() ;
         p_tdl = p_media->get_video_track_description() ;
         break ;
      /*-------------------------------------------------------------------*/
      case libvlc_track_audio :
         i_active_track = p_media->get_audio_track() ;
         p_tdl = p_media->get_audio_track_description() ;
         /*--( Track -1 (deactivate) is used to mute ... )-----------------*/
         boo_add_deactivate = false ;
         /*----------------------------------------------------------------*/
         break ;
      /*-------------------------------------------------------------------*/
      case libvlc_track_text :
         i_active_track = p_media->get_video_spu() ;
         p_tdl = p_media->get_video_spu_description() ;
         break ;
      /*-------------------------------------------------------------------*/
      default :
         p_tdl = nullptr ;
         break ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Adding track list )-----------------------------------------------*/
   int i_num_option = 0 ;
   for( auto p_desc = p_tdl ; p_desc != nullptr ;
        p_desc = p_desc->p_next, ++i_num_option
      )
   {
      /*-------------------------------------------------------------------*/
      if( boo_add_deactivate || p_desc->i_id != -1 )
      {  /*----------------------------------------------------------------*/
         menu->AppendRadioItem( i_id_start + i_num_option,
                                wxString::FromUTF8( p_desc->psz_name )
                              )->Check( p_desc->i_id == i_active_track ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( p_tdl != nullptr )
   {  p_media->release_track_description_list( p_tdl ) ; }
   /*----------------------------------------------------------------------*/
   return( menu ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylist::get_track_id_by_pos( CVlcMedia *p_media,
                                    libvlc_track_type_t type,
                                    int i_pos
                                  ) const
{
   /*----------------------------------------------------------------------*/
   libvlc_track_description_t *p_tdl ;

   /*--( Item to check in the menu )---------------------------------------*/
   switch( type )
   {
      /*-------------------------------------------------------------------*/
      case libvlc_track_video :
         p_tdl = p_media->get_video_track_description() ;
         break ;
      /*-------------------------------------------------------------------*/
      case libvlc_track_audio :
         p_tdl = p_media->get_audio_track_description() ;
         break ;
      /*-------------------------------------------------------------------*/
      case libvlc_track_text :
         p_tdl = p_media->get_video_spu_description() ;
         break ;
      /*-------------------------------------------------------------------*/
      default :
         return( -1 ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Search for the real track id. The search is done by position )----*/
   libvlc_track_description_t *p_desc = p_tdl ;
   for( ; p_desc != nullptr && i_pos > 0 ; p_desc = p_desc->p_next, --i_pos )
   {  ; }
   /*----------------------------------------------------------------------*/
   int i_track_id = ( p_desc == nullptr ? -1 : p_desc->i_id ) ;
   /*--( Free gathered data )----------------------------------------------*/
   p_media->release_track_description_list( p_tdl ) ;
   /*----------------------------------------------------------------------*/
   return( i_track_id ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxDEFINE_EVENT( EVT_AST_CONTEXT_MENU, wxCommandEvent ) ;

/*-------------------------------------------------------------------------*/
class CThreadRefreshStreams : public wxThread
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CThreadRefreshStreams( CPlaylist *p_pl, CChannel *p_channel )
                           : m_p_pl( p_pl ), m_p_channel( p_channel )
      {   ; }
      //~CThreadRefreshStreams() ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      CPlaylist *m_p_pl ;
      CChannel  *m_p_channel ;

   /*----------------------------------------------------------------------*/
   public :
      /*--( Will be executed by "Run" )------------------------------------*/
      virtual wxThread::ExitCode Entry() override
      {
         /*----------------------------------------------------------------*/
         wxCriticalSectionLocker enter_csec( m_p_pl->m_csec_context_menu ) ;
         m_p_channel->refresh_streams() ;
         /*-----------------------------------------------------------------+
         ! The event string will contain the channel name. This will enable !
         ! to fill the right contextual menu if any                         !
         +-----------------------------------------------------------------*/
         wxCommandEvent *p_event
                      = new wxCommandEvent( EVT_AST_CONTEXT_MENU ) ;
         p_event->SetString( m_p_channel->get_name() ) ;
         wxQueueEvent( m_p_pl, p_event ) ;
         /*--( Under MSW the loop has to be "woken up" ... )---------------*/
         #ifdef __WXMSW__
            SetTimer( m_p_pl->GetHandle(), 0, 0, nullptr ) ;
         #endif // __WXMSW__
         /*----------------------------------------------------------------*/
         return( ( wxThread::ExitCode )0 ) ;
         /*----------------------------------------------------------------*/
      }

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
void CPlaylist::refresh_streams( CChannel *p_channel )
{
   /*----------------------------------------------------------------------*/
   CThreadRefreshStreams *p_thread
                       = new CThreadRefreshStreams( this, p_channel ) ;
   /*----------------------------------------------------------------------*/
   p_thread->Run() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::add_streams_to_context_menu( CChannel *p_channel )
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec_menu( m_csec_context_menu ) ;
   if( m_context_menu == nullptr ) { return ; }

   /*--( If present, remove the "information" menu item )------------------*/
   wxMenuItem *mi_load = m_context_menu->FindItem( ID_PL_LOADING_STREAMS ) ;
   if( mi_load != nullptr )
   {
      /*-------------------------------------------------------------------*/
      m_context_menu->Destroy( mi_load ) ;
      /*--( Delete the separator following it )----------------------------*/
      mi_load = m_context_menu->FindItemByPosition( 0 ) ;
      if( mi_load != nullptr && mi_load->IsSeparator() )
      {  m_context_menu->Destroy( mi_load ) ; }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   bool boo_has_streams = !p_channel->is_refresh_streams_needed() ;
   /*----------------------------------------------------------------------*/
   if( !boo_has_streams && !p_channel->is_gs_retry_later() )
   {  return ; }

   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   int i_mi_pos = 0 ;
   /*----------------------------------------------------------------------*/
   mtf::insert_item( m_context_menu, i_mi_pos++, ID_PL_PLAY_ACTIVE,
                     ( boo_has_streams
                       ? _( "&Play" ) : _( "&Wait for online" )
                     ) + "\tEnter",
                     media_playback_start_16x16_xpm
                   ) ;
   /*----------------------------------------------------------------------*/
   if( frame->get_media_active() != nullptr )
   {  mtf::insert_item( m_context_menu, i_mi_pos++, ID_PL_PLAY_OTHER,
                        ( boo_has_streams
                          ? _( "Play in a &new player" )
                          : _( "Wait for online in a &new player" )
                        ) + "\tAlt+Enter",
                        media_playback_start_16x16_xpm
                      ) ;
   }

   /*----------------------------------------------------------------------*/
   m_context_menu->InsertSeparator( i_mi_pos++ ) ;

   /*-----------------------------------------------------------------------+
   ! Streams                                                                !
   +-----------------------------------------------------------------------*/
   if( boo_has_streams )
   {
      /*-------------------------------------------------------------------*/
      int i_id = ID_PL_STREAM_MIN ;
      /*-------------------------------------------------------------------*/
      for( const auto &stream : p_channel->get_vec_stream() )
      {
         /*----------------------------------------------------------------*/
         if( i_id >= ID_PL_STREAM_MAX ) { break ; }
         /*----------------------------------------------------------------*/
         m_context_menu->InsertRadioItem( i_mi_pos++, i_id++, stream.m_s_name
                                      )->Check(  p_channel->get_stream_name()
                                              == stream.m_s_name
                                              ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
      m_context_menu->InsertSeparator( i_mi_pos ) ;

      /*--( Add the separator if not already done )------------------------*/
      if( !get_browser_ch() ) { m_context_menu->AppendSeparator() ; }
      /*-------------------------------------------------------------------*/
      mtf::append_item( m_context_menu, ID_PL_PLAY_WITH_VLC,
                        _( "Play with &VLC" ),
                        mtf::get_menu_bmp( wxART_GO_FORWARD )
                      ) ;
      mtf::append_item( m_context_menu, ID_PL_COPY_URL,
                        _( "&Copy stream address" ),
                        mtf::get_menu_bmp( wxART_COPY )
                      ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::add_streams_to_context_menu( const wxString &s_channel )
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = wxGetApp().get_channel( s_channel ) ;
   if( p_channel == nullptr ) { return ; }

   /*--( The active channel must the one we are working on )---------------*/
   CChannel *p_channel_focus = get_channel( get_item_focus() ) ;
   if( p_channel_focus != p_channel ) { return ; }

   /*----------------------------------------------------------------------*/
   add_streams_to_context_menu( p_channel_focus ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylist::context_menu( wxListEvent &event )
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   int    i_item = event.GetIndex() ;
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel( i_item ) ;
   if( p_channel == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   CVlcMedia *p_media = frame->find_media( p_channel ) ;

   /*----------------------------------------------------------------------*/
   p_channel->csec_channel_enter() ;
   bool boo_has_streams = !p_channel->is_refresh_streams_needed() ;
   /*----------------------------------------------------------------------*/
   wxMenu menu ;
   /*----------------------------------------------------------------------*/
   m_csec_context_menu.Enter() ;
   m_context_menu = &menu ;

   /*--( Let some room for the streams and a few other options )-----------*/
   int i_id = ID_PL_STREAM_MAX ;
   int i_id_play_hil        = -1 ;
   int i_id_track_video_min = -1 ;
   int i_id_track_video_max = -1 ;
   int i_id_track_audio_min = -1 ;
   int i_id_track_audio_max = -1 ;
   int i_id_track_spu_min   = -1 ;
   int i_id_track_spu_max   = -1 ;
   int i_id_move_pl_num_min = -1 ;
   int i_id_move_pl_num_max = -1 ;
   int i_id_move_to_new_pl  = -1 ;
   int i_id_open_in_browser = -1 ;

   /*-----------------------------------------------------------------------+
   ! Some options are only meaningful if only one channel is hil            !
   +-----------------------------------------------------------------------*/
   if( get_nb_hil() > 1 )
   {
      /*-------------------------------------------------------------------*/
      i_id_play_hil = i_id++ ;
      mtf::append_item( &menu, i_id_play_hil,
                        _( "&Play selected channels" ),
                        media_playback_start_16x16_xpm
                      ) ;
      /*-------------------------------------------------------------------*/
   }
   else /*--( Track menu displayed if currently played )-------------------*/
   if( p_media != nullptr && boo_has_streams )
   {
      /*-------------------------------------------------------------------*/
      wxMenu *menu_all = new wxMenu ;
      wxMenu *menu_track ;
      int i_nb_track ;
      /*--( Video tracks )-------------------------------------------------*/
      i_id_track_video_min = i_id_track_video_max = i_id++ ;
      menu_track = build_track_menu( p_media, libvlc_track_video,
                                     i_id_track_video_min
                                   ) ;
      i_nb_track = menu_track->GetMenuItemCount() ;
      if( i_nb_track > 0 )
      {  i_id += i_nb_track ;
         i_id_track_video_max += i_nb_track - 1 ;
      }
      menu_all->AppendSubMenu( menu_track, _( "&Video" )
                             )->Enable( i_nb_track > 0 ) ;
      /*--( Audio tracks )-------------------------------------------------*/
      i_id_track_audio_min = i_id_track_audio_max = i_id++ ;
      menu_track = build_track_menu( p_media, libvlc_track_audio,
                                     i_id_track_audio_min
                                   ) ;
      i_nb_track = menu_track->GetMenuItemCount() ;
      if( i_nb_track > 0 )
      {  i_id += i_nb_track ;
         i_id_track_audio_max += i_nb_track - 1 ;
      }
      menu_all->AppendSubMenu( menu_track, _( "&Audio" )
                             )->Enable( i_nb_track > 0 ) ;
      /*--( Subtitles )----------------------------------------------------*/
      i_id_track_spu_min = i_id_track_spu_max = i_id++ ;
      menu_track = build_track_menu( p_media, libvlc_track_text,
                                     i_id_track_spu_min
                                   ) ;
      i_nb_track = menu_track->GetMenuItemCount() ;
      if( i_nb_track > 0 )
      {  i_id += i_nb_track ;
         i_id_track_spu_max += i_nb_track - 1 ;
      }
      menu_all->AppendSubMenu( menu_track, _( "&Text" )
                             )->Enable( i_nb_track > 0 ) ;
      /*-------------------------------------------------------------------*/
      menu.AppendSubMenu( menu_all, _( "&Tracks" ) ) ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   if( menu.GetMenuItemCount() > 0 ) { menu.AppendSeparator() ; }

   /*-----------------------------------------------------------------------+
   ! Copy/Paste                                                             !
   +-----------------------------------------------------------------------*/
   mtf::append_item( &menu, wxID_COPY, wxART_COPY ) ;
   if( get_cut_paste_ch() )
   {  mtf::append_item( &menu, wxID_CUT, wxART_CUT ) ;
      mtf::append_item( &menu, wxID_PASTE, wxART_PASTE ) ;
   }

   /*-----------------------------------------------------------------------+
   ! Rename channel                                                         !
   +-----------------------------------------------------------------------*/
   int i_id_rename_ch = -1 ;
   if( get_nb_hil() == 1 && get_rename_ch() )
   {  /*-------------------------------------------------------------------*/
      i_id_rename_ch = i_id++ ;
      mtf::append_item( &menu, i_id_rename_ch, _( "&Rename\tF2" ),
                        accessories_text_editor_16x16_xpm
                      ) ;
      /*-------------------------------------------------------------------*/
   }

   /*-----------------------------------------------------------------------+
   ! Delete from playlist                                                   !
   +-----------------------------------------------------------------------*/
   int i_id_del_ch = -1 ;
   if( get_delete_ch() )
   {  /*-------------------------------------------------------------------*/
      i_id_del_ch = i_id++ ;
      mtf::append_item( &menu, i_id_del_ch,
                        _( "&Delete from playlist" ) + '\t' + _( "Del" ),
                        mtf::get_menu_bmp( wxART_DELETE )
                      ) ;
      /*-------------------------------------------------------------------*/
   }

   /*-----------------------------------------------------------------------+
   ! Add to playlist                                                        !
   +-----------------------------------------------------------------------*/
   int i_id_add_pl_num_min = i_id ;
   wxMenu *menu_pl = new wxMenu ;
   /*--( Rename its references in the user playlists )---------------------*/
   t_vec_pl vec_pl_user ;
   frame->get_bopl()->get_pl_children( PK_USER_NODE, vec_pl_user ) ;
   /*----------------------------------------------------------------------*/
   for( const auto pl : vec_pl_user )
   {
      /*-------------------------------------------------------------------*/
      if( pl != this && pl->get_accept_ch() )
      {  menu_pl->Append( i_id, pl->get_label() ) ; }
      /*--( The id is incremented separately to access immediately the pl )*/
      ++i_id ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   int i_id_add_pl_num_max = i_id - 1 ;
   /*--( Separator added only if a user pl to "add to" is present )--------*/
   if( menu_pl->GetMenuItemCount() > 0 )
   {  menu_pl->AppendSeparator() ; }
   /*----------------------------------------------------------------------*/
   int i_id_add_to_new_pl = i_id++ ;
   mtf::append_item( menu_pl, i_id_add_to_new_pl, _( "&New playlist" ),
                     document_new_16x16_xpm, _( "Add to a new playlist" )
                   ) ;
   /*----------------------------------------------------------------------*/
   wxMenuItem *mi = menu.AppendSubMenu( menu_pl, _( "&Add to playlist" ) ) ;
   mi->SetBitmap( list_add_16x16_xpm ) ;

   /*-----------------------------------------------------------------------+
   ! Move to playlist                                                       !
   +-----------------------------------------------------------------------*/
   if( get_give_ch() )
   {
      /*-------------------------------------------------------------------*/
      i_id_move_pl_num_min = i_id ;
      menu_pl = new wxMenu ;
      /*-------------------------------------------------------------------*/
      for( const auto pl : vec_pl_user )
      {  /*----------------------------------------------------------------*/
         if( pl != this && pl->get_accept_ch() )
         {  menu_pl->Append( i_id, pl->get_label() ) ; }
         /*----------------------------------------------------------------*/
         ++i_id ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
      i_id_move_pl_num_max = i_id - 1 ;
      /*--( Separator added only if a user pl to "add to" is present )-----*/
      if( menu_pl->GetMenuItemCount() > 0 )
      {  menu_pl->AppendSeparator() ; }
      i_id_move_to_new_pl = i_id++ ;
      mtf::append_item( menu_pl, i_id_move_to_new_pl, _( "&New playlist" ),
                        document_new_16x16_xpm, _( "Move to a new playlist" )
                      ) ;
      /*-------------------------------------------------------------------*/
      mi = menu.AppendSubMenu( menu_pl, _( "&Move to playlist" ) ) ;
      mi->SetBitmap( start_here_16x16_xpm ) ;
      /*-------------------------------------------------------------------*/
   }

   /*-----------------------------------------------------------------------+
   ! Move up/down                                                           !
   +-----------------------------------------------------------------------*/
   if( get_order_ch() && i_item > 0 )
   {  mtf::append_item( &menu, ID_MOVE_UP, _( "Move up\tAlt+Up" ),
                        go_up_16x16_xpm
                      ) ;
   }
   /*----------------------------------------------------------------------*/
   if( get_order_ch() && i_item < GetItemCount() - 1 )
   {  mtf::append_item( &menu, ID_MOVE_DOWN, _( "Move down\tAlt+Down" ),
                        go_down_16x16_xpm
                      ) ;
   }
   /*----------------------------------------------------------------------*/
   if( get_nb_hil() == 1 )
   {
      /*-------------------------------------------------------------------*/
      if( get_browser_ch() )
      {  /*----------------------------------------------------------------*/
         menu.AppendSeparator() ;
         /*----------------------------------------------------------------*/
         i_id_open_in_browser = i_id++ ;
         mtf::append_item( &menu, i_id_open_in_browser,
                           _( "Open in &browser" ),
                           mtf::get_menu_bmp( wxART_EXECUTABLE_FILE )
                         ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*-----------------------------------------------------------------------+
   ! Modify menu to include the stream information or search them           !
   +-----------------------------------------------------------------------*/
   if( boo_has_streams )
   {  add_streams_to_context_menu( p_channel ) ; }
   else
   {  /*-------------------------------------------------------------------*/
      mtf::insert_item( &menu, 0, ID_PL_LOADING_STREAMS,
                        _( "Looking for channel streams ..." ),
                        emblem_important_16x16_xpm
                      )->Enable( false ) ;
      /*-------------------------------------------------------------------*/
      menu.InsertSeparator( 1 ) ;
      /*--( Searching the streams in a new thread )------------------------*/
      refresh_streams( p_channel ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( We can now "release" the locks )----------------------------------*/
   p_channel->csec_channel_leave() ;
   m_csec_context_menu.Leave() ;

   /*-----------------------------------------------------------------------+
   ! Display menu and wait for user input                                   !
   +-----------------------------------------------------------------------*/
   int i_sel_id = GetPopupMenuSelectionFromUser( menu ) ;

   /*--( The menu won't be modified outside of this function )-------------*/
   m_context_menu = nullptr ;

   /*-----------------------------------------------------------------------+
   ! Execute user choice                                                    !
   +-----------------------------------------------------------------------*/
   if( i_sel_id >= ID_PL_STREAM_MIN && i_sel_id < ID_PL_STREAM_MAX )
   {
      /*--( Change stream )------------------------------------------------*/
      p_channel->set_stream_name( i_sel_id - ID_PL_STREAM_MIN ) ;
      p_channel->set_save_cfg( true ) ;
      /*-------------------------------------------------------------------*/
      wxString s_channel = m_vs_channel[ i_item ] ;
      /*--( If the channel is currently played, it is restarted )----------*/
      for( const auto p_media : frame->get_vec_media() )
      {  /*----------------------------------------------------------------*/
         if( p_media->get_channel_name() == s_channel )
         {  frame->play_channel( s_channel ) ; }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id == ID_PL_PLAY_ACTIVE )
   {  play_item( i_item ) ; }
   else
   if( i_sel_id == ID_PL_PLAY_OTHER )
   {  play_item( i_item, true ) ; }
   else
   if( i_sel_id >= i_id_track_video_min && i_sel_id <= i_id_track_video_max )
   {  /*--( There "p_media" is not null )----------------------------------*/
      const int co_i_track
              = get_track_id_by_pos( p_media, libvlc_track_video,
                                     i_sel_id - i_id_track_video_min
                                   ) ;
      p_channel->set_video_track( co_i_track ) ;
      p_media->set_video_track( co_i_track ) ;
      p_channel->set_save_cfg( true ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id >= i_id_track_audio_min && i_sel_id <= i_id_track_audio_max )
   {  /*--( There "p_media" is not null )----------------------------------*/
      const int co_i_track
              = get_track_id_by_pos( p_media, libvlc_track_audio,
                                     i_sel_id - i_id_track_audio_min
                                   ) ;
      p_channel->set_audio_track( co_i_track ) ;
      p_media->set_audio_track( co_i_track ) ;
      p_channel->set_save_cfg( true ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id >= i_id_track_spu_min && i_sel_id <= i_id_track_spu_max )
   {  /*--( There "p_media" is not null )----------------------------------*/
      const int co_i_spu = get_track_id_by_pos( p_media, libvlc_track_text,
                                                i_sel_id - i_id_track_spu_min
                                              ) ;
      p_channel->set_video_spu( co_i_spu ) ;
      p_media->set_video_spu( co_i_spu ) ;
      p_channel->set_save_cfg( true ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id == wxID_COPY )
   {  copy_hil_to_clipboard() ; }
   else
   if( i_sel_id == wxID_CUT )
   {  cut_hil_to_clipboard() ; }
   else
   if( i_sel_id == wxID_PASTE )
   {  paste_from_clipboard() ; }
   else
   if( i_sel_id == i_id_rename_ch )
   {  ren_channel_hil() ; }
   else
   if( i_sel_id == i_id_del_ch )
   {  del_channel_hil() ; }
   else
   if( i_sel_id >= i_id_add_pl_num_min && i_sel_id <= i_id_add_pl_num_max )
   {  /*-------------------------------------------------------------------*/
      CPlaylist *pl = vec_pl_user[ i_sel_id - i_id_add_pl_num_min ] ;
      pl->add_channel_hil( this ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id == i_id_add_to_new_pl )
   {  /*-------------------------------------------------------------------*/
      CPlaylistUser *pl_new = frame->create_new_pl_user() ;
      if( pl_new != nullptr ) { pl_new->add_channel_hil( this ) ; }
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id >= i_id_move_pl_num_min && i_sel_id <= i_id_move_pl_num_max )
   {  /*-------------------------------------------------------------------*/
      CPlaylist *pl = vec_pl_user[ i_sel_id - i_id_move_pl_num_min ] ;
      pl->add_channel_hil( this ) ;
      del_channel_hil() ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id == i_id_move_to_new_pl )
   {  /*-------------------------------------------------------------------*/
      CPlaylistUser *pl_new = frame->create_new_pl_user() ;
      if( pl_new != nullptr )
      {  pl_new->add_channel_hil( this ) ; del_channel_hil() ; }
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id == ID_MOVE_UP )
   {  move_channel_hil( -1 ) ; }
   else
   if( i_sel_id == ID_MOVE_DOWN )
   {  move_channel_hil( +1 ) ; }
   else
   if( i_sel_id == i_id_open_in_browser )
   {  wxLaunchDefaultBrowser( p_channel->get_url() ) ; }
   else
   if( i_sel_id == ID_PL_PLAY_WITH_VLC )
   {
      /*-------------------------------------------------------------------*/
      const str_stream *p_stream = p_channel->get_stream() ;
      /*-------------------------------------------------------------------*/
      if( p_stream != nullptr )
      {
         /*-----------------------------------------------------------------+
         ! vlc might not be at the same place as libvlc ... especially      !
         ! under GNU/Linux.                                                 !
         ! The version of VLC is not known ... (build vlc init)             !
         +-----------------------------------------------------------------*/
         wxString s_cmd = wxGetApp().m_fn_vlc.GetFullPath()
           + ' ' + CVlcInterface::st_build_init_string( p_channel, -1 ) + ' '
           + '\'' + p_stream->get_vlc_url() + '\'' ;
         long l_pid ;

         /*----------------------------------------------------------------*/
         {  wxLogNull no_log ; // Suppress useless warnings
            l_pid = wxExecute( s_cmd ) ;
         }
         /*--( Exec KO ? (@Problem : doesn't seem to work under GNU/Linux )*/
         if( l_pid == 0 )
         {
            /*-------------------------------------------------------------*/
            wxMessageBox( _( "VLC could not be started.\n"
                             "Check the VLC path in the Preferences."
                           ),
                          _( "Error starting VLC" ), 0, frame
                        ) ;
            wxQueueEvent( frame->GetEventHandler(),
                          new wxMenuEvent( wxEVT_COMMAND_MENU_SELECTED,
                                           wxID_PREFERENCES
                                         )
                        ) ;
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   else
   if( i_sel_id == ID_PL_COPY_URL )
   {  mtf::set_clipboard( p_channel->get_stream()->get_vlc_url() ) ; }
   else
   if( i_sel_id == i_id_play_hil )
   {  /*-------------------------------------------------------------------*/
      mtf::t_vec_string vs_channel ;
      if( get_hil( vs_channel ) > 0 ) { frame->play_channel( vs_channel ) ; }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
}

/*--( A node is the sum of all its nodes )---------------------------------*/
void CPlaylistNode::init_list()
{
   /*----------------------------------------------------------------------*/
   if( !get_ti().IsOk() ) { return ; }
   /*----------------------------------------------------------------------*/
   m_vs_channel.clear() ;
   /*----------------------------------------------------------------------*/
   CBookPlaylist *bopl = wxGetApp().m_frame->get_bopl() ;
   /*----------------------------------------------------------------------*/
   wxTreeItemIdValue tiv_cookie ;
   /*----------------------------------------------------------------------*/
   for( auto ti_child = bopl->GetFirstChild( get_ti(), tiv_cookie ) ;
        ti_child.IsOk() ;
        ti_child = bopl->GetNextChild( get_ti(), tiv_cookie )
      )
   {  append_channel_from( bopl->get_playlist( ti_child ) ) ; }
   /*----------------------------------------------------------------------*/
   CPlaylist::init_list() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistNodeWithPanel::CPlaylistNodeWithPanel( wxWindow *parent,
                                                const wxString &s_name,
                                                const wxString &s_label,
                                                const wxString &s_tree_path
                                              )
                      : CPlaylistNode( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_panel = new wxPanel( parent ) ;
   m_panel->SetName( "panel_" + s_name ) ;
   Reparent( m_panel ) ;
   /*----------------------------------------------------------------------*/
   m_bsz = new wxBoxSizer( wxVERTICAL ) ;
   m_bsz->Add( this, 1, wxEXPAND ) ;
   m_panel->SetSizer( m_bsz ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistLiveNode::CPlaylistLiveNode( wxWindow *parent,
                                      const wxString &s_name,
                                      const wxString &s_label,
                                      const wxString &s_tree_path
                                    )
             : CPlaylistNodeWithPanel( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_win_onair = new CPlSubWinChOnair( m_panel, this ) ;
   m_bsz->Insert( 0, m_win_onair, 0, wxALL | wxEXPAND, 5 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistLiveNode::init_access_onair()
{
   /*----------------------------------------------------------------------*/
   if( get_win_onair() != nullptr )
   {  get_win_onair()->enable_start( !get_vs_channel().empty() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistLiveNode::init_list()
{
   /*----------------------------------------------------------------------*/
   CPlaylistNode::init_list() ;
   init_access_onair() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistUserNode::CPlaylistUserNode( wxWindow *parent,
                                      const wxString &s_name,
                                      const wxString &s_label,
                                      const wxString &s_tree_path
                                    )
             : CPlaylistNodeWithPanel( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   wxButton *bt_new_pl = new wxButton( m_panel, wxID_ANY,
                                       _( "Create a playlist" )
                                     ) ;
   bt_new_pl->SetToolTip( _( "Create a new user playlist to gather "
                             "some channels from other playlists"
                           )
                        ) ;
   /*----------------------------------------------------------------------*/
   m_bsz->Insert( 0, bt_new_pl, 0, wxALL | wxALIGN_CENTER, 5 ) ;
   /*----------------------------------------------------------------------*/
   bt_new_pl->Bind( wxEVT_BUTTON,
                    []( wxCommandEvent &event )
                    {  wxGetApp().m_frame->create_new_pl_user() ; }
                  ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistUserNode::append_channel_from( const CPlaylist *pl )
{
   /*----------------------------------------------------------------------*/
   if( pl == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   size_t sz_start = get_nb_channel() ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : pl->get_vs_channel() )
   {
      /*--------------------------------------------------------------------+
      ! Duplicates are avoided. Computation is used because iterators can   !
      ! be invalidated if the vector is reallocated during its growth.      !
      ! "end()" could have been used but this would have generated useless  !
      ! computations                                                        !
      +--------------------------------------------------------------------*/
      mtf::t_vec_string_it it_end = m_vs_channel.begin() + sz_start ;
      if( std::find( m_vs_channel.begin(), it_end, s_channel ) == it_end )
      {  m_vs_channel.emplace_back( s_channel ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--( A filter is supposed to filter its parent channels )-----------------*/
void CPlaylistFilter::init_list()
{
   /*----------------------------------------------------------------------*/
   m_vs_channel.clear() ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : m_pl_src->get_vs_channel() )
   {  /*-------------------------------------------------------------------*/
      CChannel *p_channel = wxGetApp().get_channel( s_channel ) ;
      if( p_channel != nullptr && m_fnc_filter( p_channel )  )
      {  m_vs_channel.emplace_back( s_channel ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   CPlaylist::init_list() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistSearch::CPlaylistSearch( wxWindow *parent,
                                  const wxString &s_name,
                                  const wxString &s_label,
                                  const wxString &s_tree_path
                                )
               : CPlaylistWithPanel( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_boo_give_ch   = true  ;
   m_boo_delete_ch = true  ;
   m_boo_order_ch  = true  ;
   /*----------------------------------------------------------------------*/
   m_search = new wxSearchCtrl( m_panel, wxID_ANY, wxEmptyString,
                                wxDefaultPosition, wxDefaultSize,
                                wxTE_PROCESS_ENTER | wxBORDER_DEFAULT
                              ) ;
   /*--( A bug keeps the "light grey" color for text entered by the user )-*/
   //m_search->SetDescriptiveText( _( "Search number or part of name" ) ) ;
   m_search->ShowCancelButton( true ) ;
   /*----------------------------------------------------------------------*/
   m_bsz->Insert( 0, m_search, 0, wxEXPAND ) ;
   /*----------------------------------------------------------------------*/
   m_search->Bind( wxEVT_TEXT_ENTER,
                   [ this ]( wxCommandEvent &event ) { exec_search() ; }
                 ) ;
   m_search->Bind( wxEVT_SEARCHCTRL_SEARCH_BTN,
                   [ this ]( wxCommandEvent &event ) { exec_search() ; }
                 ) ;
   m_search->Bind( wxEVT_SEARCHCTRL_CANCEL_BTN,
                   [ this ]( wxCommandEvent &event )
                   {  m_search->Clear() ; init_list() ; }
                 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistSearch::exec_search()
{
   /*----------------------------------------------------------------------*/
   wxString s_search = m_search->GetValue().Trim( false ).Trim() ;
   /*----------------------------------------------------------------------*/
   m_vs_channel.clear() ;
   /*----------------------------------------------------------------------*/
   if( !s_search.empty() )
   {
      /*-------------------------------------------------------------------*/
      long l_search_num ;
      if( !s_search.ToLong( &l_search_num ) ) { l_search_num = -1 ; }
      /*-------------------------------------------------------------------*/
      s_search = mtf::prepare_string( s_search ) ;
      /*-------------------------------------------------------------------*/
      for( auto it : wxGetApp().m_map_channel )
      {
         /*----------------------------------------------------------------*/
         CChannel *p_channel = it.second ;
         /*--( Name or channel number )------------------------------------*/
         if(    p_channel->get_name_find().Find( s_search ) != wxNOT_FOUND
             || (    l_search_num > 0
                  && p_channel->is_num_set()
                  && p_channel->get_num() == l_search_num
                )
           )
         {  m_vs_channel.emplace_back( p_channel->get_name() ) ; }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   CPlaylist::init_list() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistWithPanel::CPlaylistWithPanel( wxWindow *parent,
                                        const wxString &s_name,
                                        const wxString &s_label,
                                        const wxString &s_tree_path
                                      )
                  : CPlaylist( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_panel = new wxPanel( parent ) ;
   m_panel->SetName( "panel_" + s_name ) ;
   Reparent( m_panel ) ;
   /*----------------------------------------------------------------------*/
   m_bsz = new wxBoxSizer( wxVERTICAL ) ;
   m_bsz->Add( this, 1, wxEXPAND ) ;
   m_panel->SetSizer( m_bsz ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistUser::CPlaylistUser( wxWindow *parent,
                              const wxString &s_name,
                              const wxString &s_label,
                              const wxString &s_tree_path
                            )
             : CPlaylistWithPanel( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_boo_accept_ch    = true ;
   m_boo_give_ch      = true ;
   m_boo_cut_paste_ch = true ;
   m_boo_delete_ch    = true ;
   m_boo_order_ch     = true ;
   m_boo_deletable    = true ;
   m_boo_renamable    = true ;
   m_boo_movable      = true ;
   /*----------------------------------------------------------------------*/
   m_banner = new wxBannerWindow( m_panel, wxTOP ) ;
   m_banner->SetText( _( "To add channels :" ),
                      _( "- Right click on them and \"add\" them\n"
                         "- Drag and drop them\n"
                         "- Copy/Paste them"
                       )
                    ) ;
   const wxColour co_colour_bk = m_panel->GetBackgroundColour() ;
   m_banner->SetGradient( co_colour_bk, co_colour_bk ) ;
   /*----------------------------------------------------------------------*/
   m_bsz->Insert( 0, m_banner, 0, wxEXPAND ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistUser::init_list()
{
   /*----------------------------------------------------------------------*/
   CPlaylist::init_list() ;
   /*--( Display info when the list is empty )-----------------------------*/
   if( m_banner->IsShown() != m_vs_channel.empty() )
   {
      /*-------------------------------------------------------------------*/
      m_banner->Show( m_vs_channel.empty() ) ;
      m_bsz->Layout() ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistFbtvoOk::CPlaylistFbtvoOk( wxWindow *parent,
                                    const wxString &s_name,
                                    const wxString &s_label,
                                    const wxString &s_tree_path
                                  )
                : CPlaylistWithPanel( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_boo_text_long = false ; // false because all channels are from fbtv
   m_boo_give_ch   = true  ;
   m_boo_delete_ch = true  ;
   m_boo_order_ch  = true  ;
   /*----------------------------------------------------------------------*/
   m_bsz->Insert( 0, new CPlSubWinFbtvoOk( m_panel, this ), 0,
                  wxALL | wxEXPAND, 5
                ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlaylistLive::CPlaylistLive( wxWindow *parent,
                              const wxString &s_name,
                              const wxString &s_label,
                              const wxString &s_tree_path
                            )
             : CPlaylistWithPanel( parent, s_name, s_label, s_tree_path )
{
   /*----------------------------------------------------------------------*/
   m_boo_text_long  = false ; // Channels listed belong to this pl
   m_boo_rename_ch  = true  ;
   m_boo_delete_ch  = true  ;
   m_boo_order_ch   = true  ;
   m_boo_browser_ch = true  ;
   /*----------------------------------------------------------------------*/
   m_win_action = new CPlSubWinLive( m_panel, this ) ;
   m_bsz->Insert( 0, m_win_action, 0, wxALL | wxEXPAND ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CPlSubWinChOnair *CPlaylistLive::get_win_onair() const
{  return( m_win_action->get_win_onair() ) ; }

/*-------------------------------------------------------------------------*/
CChannel *CPlaylistLive::find_channel( const wxString &s_id ) const
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_source = get_source() ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : m_vs_channel )
   {
      /*-------------------------------------------------------------------*/
      CChannelLive *p_channel
      = static_cast< CChannelLive * >( wxGetApp().get_channel( s_channel ) );
      /*-------------------------------------------------------------------*/
      if(    p_channel != nullptr
          && p_channel->get_pl_source() == co_s_source
          && p_channel->get_id() == s_id
        )
      {  return( p_channel ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( nullptr ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistLive::string_to_datas( const wxString &s_channel_list )
{
   /*----------------------------------------------------------------------*/
   wxString s_origin ;
   wxString s_id ;
   wxString s_name ;
   wxStringTokenizer tkz( s_channel_list, '\n' ) ;
   /*----------------------------------------------------------------------*/
   while( tkz.HasMoreTokens() )
   {
      /*-------------------------------------------------------------------*/
      if( mtf::match( tkz.GetNextToken(), R"=(^(.*?)\t(.*)$)=", &s_id,&s_name
                    ) == 3
        )
      {  /*----------------------------------------------------------------*/
         CChannelLive *p_channel = new_channel( s_id, s_name ) ;
         /*--( Add the channel to the main structure )---------------------*/
         wxGetApp().add_channel( p_channel ) ;
         /*--( Add it to the current playlist )----------------------------*/
         m_vs_channel.emplace_back( p_channel->get_name() ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CPlaylistLive::datas_to_string() const
{
   /*----------------------------------------------------------------------*/
   wxString s_res ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_channel : m_vs_channel )
   {
      /*-------------------------------------------------------------------*/
      if( !s_res.empty() ) { s_res += '\n' ; }
      /*-------------------------------------------------------------------*/
      const CChannelLive *p_channel
      = static_cast< CChannelLive * >( wxGetApp().get_channel( s_channel ) );
      /*-------------------------------------------------------------------*/
      s_res += p_channel->get_id() + '\t' + p_channel->get_short_name() ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( s_res ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistLive::init_access_onair()
{
   /*----------------------------------------------------------------------*/
   if( get_win_action() != nullptr )
   {  get_win_action()->enable_onair( !get_vs_channel().empty() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylistLive::add_channel( CChannelLive *p_channel )
{
   /*----------------------------------------------------------------------*/
   if( p_channel == nullptr ) { return( -1 ) ; }
   /*--( Add the channel to the main structure )---------------------------*/
   wxGetApp().add_channel( p_channel ) ;
   /*--( Add it to the current playlist )----------------------------------*/
   m_vs_channel.emplace_back( p_channel->get_name() ) ;
   /*----------------------------------------------------------------------*/
   set_save_cfg( true ) ;
   /*----------------------------------------------------------------------*/
   init_list() ;
   wxGetApp().m_frame->get_bopl()->init_list_parents( this ) ;
   /*--( Focus the new channel )-------------------------------------------*/
   set_item_active( get_nb_channel() - 1 ) ;
   /*----------------------------------------------------------------------*/
   init_access_onair() ;
   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CPlaylistLive::del_channel_hil()
{
   /*----------------------------------------------------------------------*/
   if( get_nb_hil() == 0 ) { return( 0 ) ; }
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   mtf::t_vec_string vs_channel ; // Channel names
   get_hil( vs_channel ) ;
   /*--( Close media playing these source channels )-----------------------*/
   for( const auto &s_channel : vs_channel )
   {  /*-------------------------------------------------------------------*/
      CVlcMedia *p_media = frame->find_media( s_channel ) ;
      if( p_media != nullptr ) { frame->close_media( p_media, true ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*--( Delete entries from local vector )--------------------------------*/
   del_channel( vs_channel ) ;
   /*--( Delete from global channel vector )-------------------------------*/
   int i_nb_deleted = wxGetApp().del_channel( vs_channel ) ;
   if( i_nb_deleted <= 0 ) { return( -1 ) ; }
   /*----------------------------------------------------------------------*/
   set_save_cfg( true ) ;
   /*----------------------------------------------------------------------*/
   init_list() ;
   frame->get_bopl()->init_list_parents( this ) ;
   /*--( Clean playlists referencing the deleted channels )----------------*/
   frame->get_bopl()->del_channel_pl( vs_channel ) ;

   /*--( The channels will disappear : keep only one selected )------------*/
   if( GetItemCount() > 0 )
   {  set_item_active( get_item_focus() ) ; }
   /*----------------------------------------------------------------------*/
   init_access_onair() ;
   /*----------------------------------------------------------------------*/
   return( i_nb_deleted ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CPlaylistLive::add_streams_to_context_menu( CChannel *p_channel )
{
   /*-----------------------------------------------------------------------+
   ! For now only "live" channels have dynamics streams.                    !
   ! This is why this function has been created.                            !
   +-----------------------------------------------------------------------*/
   p_channel->csec_channel_enter() ;
   CPlaylist::add_streams_to_context_menu( p_channel ) ;
   p_channel->csec_channel_leave() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                        End of file CPlaylist.cpp                          !
+==========================================================================*/
